# *- coding: utf-8 -*-
"""
Utilities and classes to deal with model transformations

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext

from future.utils import with_metaclass

from . import jinja2
from . import utils
from .classbuilder import ProtocolBase
from .schema_metaclass import SchemaMetaclass

_ = gettext.gettext

# loader to register module with a transforms folder where to look for model transformations
transforms_module_loader = utils.GenericModuleFileLoader('transforms')


def _process_transform(string):
    if utils.is_pattern(string):
        return jinja2.templatedString(string)
    else:
        return utils.import_from_string(string)


class ObjectTransform(with_metaclass(SchemaMetaclass, ProtocolBase)):
    """
    Class to do simple model to model transformation
    """

    schemaUri = "http://numengo.org/ngoschema/object_transform"

    def __init__(self, **kwargs):
        ProtocolBase.__init__(self, **kwargs)
        try:
            self._from = utils.import_from_string(str(self['from']))
        except Exception as er:
            self._from = None
        try:
            self._to = utils.import_from_string(str(self['to']))
        except Exception as er:
            self._to = None

        complexTransformFrom = kwargs.get("complexTransformFrom", {})
        self._complexTransformFrom = {
            k: _process_transform(v)
            for k, v in complexTransformFrom.items()
        }

        complexTransformTo = kwargs.get("complexTransformTo", {})
        self._complexTransformTo = {
            k: _process_transform(v)
            for k, v in complexTransformTo.items()
        }

    def transform_from(self, from_, **opts):
        """
        Create a dict/object from another one using the translation dictionaries
        defined in ObjectTransform, applying the fieldsEquivalence translation
        table , and the complexTransformFrom translation dictionary

        :param from_: object/dict to transform
        :type from_: object
        :param opts: dictionary of options, objectClass=None
        :rtype: object
        """
        from_ = from_.as_dict() if hasattr(from_, "as_dict") else from_
        to_ = {}
        from_to = {v.for_json(): k for k, v in self.fieldsEquivalence.items()}
        for k, v in from_.items():
            if k in from_to:
                k2 = from_to[k]
                to_[k2] = v
        for k2, tf in self._complexTransformFrom.items():
            try:
                if isinstance(tf, jinja2.templatedString):
                    to_[k2] = tf({"to": to_, "from": from_})
                else:
                    to_[k2] = tf(**from_)
            except Exception:
                pass

        # converting types than can be easily converted by evaluation
        if self._to and hasattr(self._to, "as_dict"):
            for k, v in to_.items():
                try:
                    if self._to.__propinfo__[k]["type"] != "string":
                        to_[k] = eval(v)
                except Exception:
                    pass

            if "objectClass" not in opts:
                opts["objectClass"] = self._to

        return utils.process_collection(to_, **opts)

    def transform_to(self, to_, **opts):
        """
        Create a dict/object from another one using the translation dictionaries
        defined in ObjectTransform, applying the fieldsEquivalence translation
        table (inverted), and the complexTransformTo translation dictionary

        :param from_: object/dict to transform
        :type from_: object
        :param opts: dictionary of options, objectClass=None
        :rtype: object
        """
        to_ = to_.as_dict() if hasattr(to_, "as_dict") else to_
        from_ = {}
        to_from = {k: v.for_json() for k, v in self.fieldsEquivalence.items()}
        for k, v in to_.items():
            if k in to_from:
                k2 = to_from[k]
                from_[k2] = v
        for k2, tf in self._complexTransformTo.items():
            try:
                if isinstance(tf, jinja2.templatedString):
                    from_[k2] = tf({"to": to_, "from": from_})
                else:
                    from_[k2] = tf(**to_)
            except Exception:
                pass

        # converting types than can be easily converted by evaluation
        if self._from and hasattr(self._from, "as_dict"):
            for k, v in from_.items():
                try:
                    if self._from.__propinfo__[k]["type"] != "string":
                        from_[k] = eval(v)
                except Exception:
                    pass

            if "objectClass" not in opts:
                opts["objectClass"] = self._from

        return utils.process_collection(from_, **opts)

    def transform(self, from_, **opts):
        if hasattr(from_, "as_dict"):
            if isinstance(from_, self._from):
                return self.transform_from(from_, **opts)
            if isinstance(from_, self._to):
                return self.transform_to(from_, **opts)
        if 'objectClass' in opts:
            oc = opts['objectClass']
            if issubclass(oc, self._to):
                return self.transform_from(from_, **opts)
            if issubclass(oc, self._from):
                return self.transform_to(from_, **opts)
        raise ValueError('cannot determine which transformation to do.')



# *- coding: utf-8 -*-
"""
de/serializer meta class

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import io
import json
import logging
from abc import ABCMeta
from abc import abstractmethod
from builtins import str

import six
from future.utils import with_metaclass
from ruamel import yaml
from ruamel.yaml import YAML

from . import decorators
from . import utils

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

_ = gettext.gettext


class Serializer(with_metaclass(ABCMeta)):
    logger = logging.getLogger(__name__)

    def dump(self, obj, path, overwrite=False, **opts):
        """
        Serialize an object to a file like object string

        :param obj: object to serialize
        :param path: file path containing the object
        :type path: path
        :param overwrite: overwrites existing file
        :param opts: dictionary of option, as protocol(=w) , encoding=(utf-8)
        """
        ptcl = opts.get("protocol", "w")
        enc = opts.get("encoding", "utf-8")

        self.logger.info(_("DUMP %r to file %s" % (obj, path)))

        if path.exists() and not overwrite:
            raise IOError(_("file %s already exists" % str(path)))
        with io.open(str(path), ptcl, encoding=enc) as outfile:
            stream = self.dumps(obj, **opts)
            stream = six.text_type(stream)
            outfile.write(stream)

    @abstractmethod
    def dumps(self, obj, **opts):
        """
        Serialize an object to a string

        :param obj: object to serialize
        :param opts: dictionary of options, as protocol(=w) , encoding=(utf-8), objectClass=None
        """


serializer_registry = utils.GenericRegistry()


@serializer_registry.register()
class JsonSerializer(Serializer):
    """
    Default settings for dumping are:
    indent: 2
    ensure_ascii: False
    """

    logger = logging.getLogger(__name__ + ".JsonDeserializer")

    def dumps(self, obj, **opts):
        data = obj.as_dict() if hasattr(obj, "as_dict") else obj
        data = utils.process_collection(data, **opts)
        return json.dumps(
            data,
            indent=opts.get("indent", 2),
            ensure_ascii=opts.get("ensure_ascii", False),
            separators=opts.get("separators", None),
            # encoding=opts.get('encoding','utf-8'),
            default=opts.get("default", None),
        )


@serializer_registry.register()
class YamlSerializer(Serializer):
    logger = logging.getLogger(__name__ + ".YamlDeserializer")

    def __init__(self, **kwargs):
        # default, if not specfied, is 'rt' (round-trip)
        self._yaml = YAML(typ="safe", **kwargs)
        self._yaml = yaml

    def dumps(self, obj, **opts):
        self._yaml.indent = opts.get("indent", 2)
        self._yaml.allow_unicode = opts.get("encoding", "utf-8") == "utf-8"

        data = obj.as_dict() if hasattr(obj, "as_dict") else obj
        data = utils.process_collection(data, **opts)
        output = StringIO()
        self._yaml.safe_dump(data, output, default_flow_style=False)
        return output.getvalue()



# *- coding: utf-8 -*-
"""
Utilities and classes to deal with validators and resolvers

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import imp
import json
import logging
import pkgutil
from builtins import str

import inflection
import six
from jsonschema._utils import URIDict
from ngofile.list_files import list_files

from .exceptions import SchemaError

_ = gettext.gettext

_all_schemas_store = URIDict()


def get_all_schemas_store():
    return _all_schemas_store


def _id_of(schema):
    return schema.get("$id", schema.get("id"))


def _load_schema(name):
    """
    Load a schema from ./schemas/``name``.json and return it.

    """
    data = pkgutil.get_data("ngoschema", "schemas/{0}.json".format(name))
    #return json.loads(data.decode("utf-8"), object_pairs_hook=collections.OrderedDict)
    return json.loads(data.decode("utf-8"))


def load_schema(schema, schemas_store=None):
    """
    Load a schema to the schema store.

    if no schema store is provided, only the internal schema store is filled.

    :param schema: schema as dictionary
    :type schema: dict
    :param schemas_store: optional schemas_store to fill 
    :type schemas_store: dict
    """
    uri = _id_of(schema).rstrip('#')
    if not uri and "title" in schema:
        uri = inflection.parameterize(six.text_type(schema["title"]), "_")
    if not uri:
        raise SchemaError(
            _("Impossible to load schema because `id (or `$id) and `title fields"
              "are missing.\n%s" % schema))
    if schemas_store is not None:
        schemas_store[uri] = schema
    _all_schemas_store[uri] = schema
    return uri, schema


def load_schema_file(schema_path, schemas_store=None):
    """
    Load a schema from a file to the metaschema store
    and returns the schema dictionary.
    
    :param schema_path: path to file containing schema
    :type schema_path: [string, path]
    :param schemas_store: optional schemas_store to fill 
    :type schemas_store: dict
    :rtype: dict
    """
    with open(str(schema_path), "rb") as f:
        #schema = json.loads(f.read().decode("utf-8"), object_pairs_hook=collections.OrderedDict)
        schema = json.loads(f.read().decode("utf-8"))
        return load_schema(schema, schemas_store)


def load_module_schemas(module="ngoschema", schemas_store=None):
    """
    Load the schemas of a module that are in the folder module
    as $(MODULEPATH)/schemas/*.json and add them with load_chema_file.
    User can provide an existing schema store to fill, or a new one
    will be created.

    return the loaded schema store
    
    :param module: module name where to look
    :param schemas_store: optional schemas_store to fill 
    :type schemas_store: dict
    :rtype: dict
    """
    logger = logging.getLogger(__name__)
    libpath = imp.find_module(module)[1]

    if schemas_store is None:
        schemas_store = URIDict()
    for ms in list_files(libpath, "schemas/*.json", recursive=True):
        try:
            load_schema_file(ms, schemas_store)
        except Exception as er:
            logger.error(_("Impossible to load file %s." % ms))
            logger.error(_("%s" % er))
    return schemas_store



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import re
import sys
from builtins import object

from .str_utils import multiple_replace

# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

##### C. ROMAN (NUMENGO) modifies regex to retrieve type of parameters

PARAM_OR_RETURNS_REGEX = re.compile(":(?:param|type|rtype)")
RETURNS_REGEX = re.compile(":rtype:\s*(?P<doc>.*)", re.S)
PARAM_REGEX = re.compile(
    ":param (?P<name>[\*\w]+):\s*(?P<doc>.*?)"
    "(?:(?=:param)|(?=:type)|(?=:rtype)|(?=:raises)|\Z)",
    re.S,
)
TYPE_REGEX = re.compile(
    ":type (?P<name>[\*\w]+):\s*(?P<type>.*?)"
    "(?:(?=:param)|(?=:type)|(?=:rtype)|(?=:raises)|\Z)",
    re.S,
)


def trim(docstring):
    """trim function from PEP-257"""
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Current code/unittests expects a line return at
    # end of multiline docstrings
    # workaround expected behavior from unittests
    if "\n" in docstring:
        trimmed.append("")

    # Return a single string:
    return "\n".join(trimmed)


def reindent(string):
    return "\n".join(l.strip() for l in string.strip().split("\n"))


def parse_docstring(docstring):
    """
    Parse the docstring into its components.

    :param docstring: docstring to parse
    :returns: a dictionary of form
              {
                  "shortDescription": ...,
                  "longDescription": ...,
                  "params": { name: {"type", ..., "doc": ...}},
                  "returns": ...
              }
    """

    short_description = long_description = returns = ""
    params = []

    def add2dict(dic, k, el):
        k = k.strip()
        if k not in dic:
            dic[k] = {}
        dic[k].update(el)

    if docstring:
        docstring = trim(docstring)

        lines = docstring.split("\n", 1)
        short_description = lines[0]

        if len(lines) > 1:
            long_description = lines[1].strip()

            params_returns_desc = None

            match = PARAM_OR_RETURNS_REGEX.search(long_description)
            if match:
                long_desc_end = match.start()
                params_returns_desc = long_description[long_desc_end:].strip()
                long_description = long_description[:long_desc_end].rstrip()

            if params_returns_desc:
                all = PARAM_REGEX.findall(params_returns_desc)
                params = {
                    name.strip(): {
                        "doc": trim(doc).strip()
                    }
                    for name, doc in PARAM_REGEX.findall(params_returns_desc)
                }
                types = TYPE_REGEX.findall(params_returns_desc)
                for name, typ in types:
                    add2dict(params, name, {"type": typ.strip()})

                match = RETURNS_REGEX.search(params_returns_desc)
                if match:
                    returns = reindent(match.group("doc"))

    return {
        "doc": docstring,
        "shortDescription": short_description,
        "longDescription": long_description,
        "params": params,
        "returns": returns,
    }


class InfoMixin(object):
    @classmethod
    def _get_doc(cls):
        """
        Return documentary of class
        By default it returns docstring of class, but it can be overridden
        for example for cases like merging own docstring with parent
        """
        return cls.__doc__

    @classmethod
    def get_info(cls):
        doc = parse_docstring(cls._get_doc())

        return {
            "name": cls.get_name(),
            "platform": cls.get_platform(),
            "module": cls.__module__,
            "title": doc["shortDescription"],
            "description": doc["longDescription"],
            "parameters": doc["params"],
            "returns": doc["returns"],
        }


def parse_type_string(typestring):
    """
    Parse a type string and returns a json schema compliant type
    """
    # this is not pretty, but it should cover most cases
    typestring = typestring.strip()
    if typestring[0] != "{":
        if typestring.startswith("enum"):
            typestring = "{%s}" % typestring
        else:
            typestring = "{type:%s}" % typestring
    ret = re.sub(r"(\w+)", r'"\1"', typestring)
    aliases = {
        "=": ":",
        '"tuple"': '"list"',
        '"set"': '"list", "uniqueItems":True',
        '"bool"': '"boolean"',
        '"int"': '"integer"',
        '"float"': '"number"',
        '"str"': '"string"',
        '"strArray"': '"array", "items": {"type": "string"}',
        '"stringArray"': '"array", "items": {"type": "string"}',
        '"text"': '"string"',
        '"dict"': '"object"',
        '"choice"': '"enum"',
        '"uri"-"reference"': '"uri-reference"',
        '"uri"-"template"': '"uri-template"',
        '"date"-"time"': '"date-time"',
        '"json"-"pointer"': '"json-pointer"',
        '"pathlib.Path"': '"path"',
        '"True"': "True",
        '"true"': "True",
        '"False"': "False",
        '"false"': "False",
    }
    ret = multiple_replace(ret, aliases)
    return eval(ret)



# -*- coding: utf-8 -*-
"""

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import json
import os.path
import logging

from jsonschema import RefResolver

from ngoschema import DEFAULT_MS_URI
from ngoschema.resolver import ExpandingResolver
from ngoschema.resolver import get_resolver
from ngoschema.schemas_loader import get_all_schemas_store
from ngoschema.schemas_loader import load_schema_file

logging.basicConfig(level=logging.INFO)

test_dir = os.path.dirname(os.path.realpath(__file__))
defs_fp = os.path.join(test_dir, "schemas", "ngo-defs-draft-XX.json")
DEFS_URI, DEFS = load_schema_file(defs_fp)
MS_STORE = get_all_schemas_store()

resolver = ExpandingResolver(DEFAULT_MS_URI, MS_STORE[DEFAULT_MS_URI],
                             MS_STORE)
orig_resolver = RefResolver(DEFAULT_MS_URI, MS_STORE[DEFAULT_MS_URI], MS_STORE)


def test_resolver():
    ref1 = "%s#/definitions/Metadata" % DEFAULT_MS_URI
    ref2 = "#/definitions/Metadata"
    assert resolver.resolve(ref1) == resolver.resolve(ref2)
    assert resolver.resolve("http://json-schema.org/geo")


def test_resolve_by_name():
    id, sch = resolver.resolve_by_name("RealVariable")
    assert sch


def test_expand_schema():
    id, sch = resolver.resolve_by_name("ComponentDefinition")
    id, sch2 = orig_resolver.resolve(id)

    if False:
        with open("ComponentDefinition_expanded.json", "w") as outfile:
            json.dump(sch, outfile, indent=2)

        with open("ComponentDefinition.json", "w") as outfile:
            json.dump(sch2, outfile, indent=2)

    assert len(sch["properties"]) > len(sch2["properties"])
    # pprint(sch2)


def test_get_resolver():
    resolver = get_resolver()
    id, sch = resolver.resolve_by_name("ComponentDefinition")


if __name__ == "__main__":
    test_resolver()
    test_resolve_by_name()
    test_expand_schema()
    test_get_resolver()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import logging
import os.path

from ngoschema.object_loader import ObjectLoader
from ngoschema.config import ConfigLoader
from ngoschema.classbuilder import objects_config_loader

_ = gettext.gettext

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("python_jsonschema_objects.classbuilder").setLevel(
    logging.INFO)

dirpath = os.path.dirname(os.path.realpath(__file__))
cfgfile = os.path.join(dirpath, "objects", "config.ini")


def test_config():
    cm = ConfigLoader()
    cm.add_config(cfgfile)
    assert cm.section("Section:subsection")["OnlyInSection"] == "yes"
    assert cm.section("Section:subsection")["InSubsection"] == "yes"


def test_object_config_loader():
    objects_config_loader.add_config(cfgfile)
    prj = ObjectLoader()
    prj.set_configfiles_defaults()
    assert prj.primaryKey == "NotTheDefaultOne"


if __name__ == "__main__":
    test_config()
    test_object_config_loader()



# *- coding: utf-8 -*-
"""
Utilities for learning about live objects, including modules, classes, instances,
functions, and methods

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 02/01/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import ast
import gettext
import importlib
import inspect
import itertools
import logging
from builtins import next
from builtins import object
from builtins import str
from builtins import zip

import six

from .doc_rest_parser import parse_docstring
from .doc_rest_parser import parse_type_string
from .exceptions import InvalidValue
from .utils import import_from_string
from .utils import is_string

_ = gettext.gettext

logger = logging.getLogger(__name__)

EXCLUDED_MODULES = [
    None, str.__class__.__module__, "future.types.newobject",
    "__builtin__", '_abcoll', 'abc'
]

class Argument(object):
    """
    Class for function/method attributes
    """

    def __init__(self, name, doc=None, doctype=None):
        self.name = name
        self._has_default = False
        self._doc = doc
        self._type = None
        if doctype:
            try:
                self._type = parse_type_string(doctype)
            except Exception as er:
                logger.error(
                    "impossible to parse valid schema from type doc %s" %
                    doctype)

    def __repr__(self):
        ret = "<arg %s" % self.name
        if self._has_default:
            ret += "=%s" % self.default
        return ret + ">"

    @property
    def doc(self):
        return self._doc

    @property
    def type(self):
        return self._type

    @property
    def default(self):
        if self._has_default:
            return self._default
        return

    @default.setter
    def default(self, value):
        self._has_default = True
        self._default = value


# global variale used in visit_FunctionDef
# set by inspectors when inspecting a class/function
_module = None


def _id(arg):
    return getattr(arg, "id", None) or getattr(arg, "arg")


def visit_FunctionDef(node):
    """ ast node visitor """
    doc = parse_docstring(ast.get_docstring(node))
    _short_desc = doc["shortDescription"]
    _long_desc = doc["longDescription"]

    doc_params = doc["params"]
    _returns = doc["returns"]

    docs = [
        doc_params[_id(a)].get("doc", None) if _id(a) in doc_params else None
        for a in node.args.args
    ]
    doctypes = [
        doc_params[_id(a)]["type"]
        if _id(a) in doc_params and "type" in doc_params[_id(a)] else None
        for a in node.args.args
    ]

    _params = [
        Argument(_id(arg), doc, doctype)
        for arg, doc, doctype in zip(node.args.args, docs, doctypes)
    ]

    _keywords = node.args.kwarg.arg if hasattr(node.args.kwarg,
                                               'arg') else node.args.kwarg
    _varargs = node.args.vararg.arg if hasattr(node.args.vararg,
                                               'arg') else node.args.vararg
    _defaults = [ast.literal_eval(d) for d in node.args.defaults]
    for d, p in zip(reversed(_defaults), reversed(_params)):
        p.default = d

    decorators = []
    for n in node.decorator_list:
        name = ""
        if isinstance(n, ast.Call):
            name = n.func.attr if isinstance(n.func,
                                             ast.Attribute) else _id(n.func)
            args = [
                d.attr if isinstance(d, ast.Attribute) else ast.literal_eval(d)
                for d in n.args
            ]
        else:
            name = n.attr if isinstance(n, ast.Attribute) else _id(n)
            args = []
        dec = getattr(_module, name, None)
        if not dec:
            dec = name
        else:
            dec = FunctionInspector(dec)
            for a, p in zip(args, dec.parameters):
                p.default = a
        decorators.append(dec)
    return _short_desc, _long_desc, _returns, _params, _keywords, _varargs, decorators


class FunctionInspector(object):
    """
    Class to inspect a function
    """

    def __init__(self, function=None, **kwargs):
        """
        FunctionInspector is normally initialized giving it a function
        """
        self.name = None
        self.doc = None
        self.shortDescription = None
        self.longDescription = None
        self.returns = None
        self.parameters = []
        self.keywords = None
        self.varargs = None
        self.decorators = []

        for k, v in list(kwargs.items()):
            setattr(self, k, v)
        if function:
            self._from_function(function)

    def has_param(self, param):
        return any([p.name == param for p in self.parameters])

    def _from_function(self, function):
        global _module
        if is_string(function):
            function = import_from_string(function)
        if not (inspect.isfunction(function) or inspect.ismethod(function)
                or type(function) in [staticmethod, classmethod]):
            raise InvalidValue(
                _("%r is not a function or a method" % function))
        self.function = function
        self.doc = self.function.__doc__
        if self.doc:
            self.doc = self.doc.strip()

        if not type(function) in [staticmethod, classmethod]:
            self.name = getattr(function, "__name__", None)
            _module = importlib.import_module(function.__module__)
            self.module = _module
            self.moduleName = _module.__name__
            if self.moduleName in EXCLUDED_MODULES:
                return

        def _visit_FunctionDef(node):
            if (node.name == self.name
                    and node.lineno == self.function.__code__.co_firstlineno):
                sd, ld, rt, ps, kw, va, decs = visit_FunctionDef(node)
                self.shortDescription = sd
                self.longDescription = ld
                self.returns = rt
                self.parameters = ps
                self.keywords = kw
                self.varargs = va
                self.decorators = decs

        node_iter = ast.NodeVisitor()
        node_iter.visit_FunctionDef = _visit_FunctionDef
        node_iter.visit(ast.parse(inspect.getsource(_module)))

    def __repr__(self):
        return "<FunctionInsp %s>" % self.name.__repr__()


class ClassInspector(object):
    """
    Class to inspect a class
    """

    def __init__(self, klass):
        self.name = None
        self.doc = None
        self.shortDescription = None
        self.longDescription = None
        self.attributes = {}
        self.attributesInherited = {}
        self.properties = {}
        self.propertiesInherited = {}
        self.methods = {}
        self.methodsInherited = {}
        self.mro = []
        self._from_class(klass)

    def _from_class(self, klass):
        global _module
        if is_string(klass):
            klass = import_from_string(klass)
        if not inspect.isclass(klass):
            raise InvalidValue(_("%r is not a class" % klass))

        self.klass = klass
        self.name = klass.__name__
        self.doc = klass.__doc__
        if self.doc:
            self.doc = self.doc.strip()
            doc = parse_docstring(klass.__doc__)
            self.shortDescription = doc["shortDescription"]
            self.longDescription = doc["longDescription"]
        else:
            self.shortDescription = None
            self.longDescription = None
        _module = importlib.import_module(klass.__module__)
        self.module = _module
        self.moduleName = _module.__name__

        ms = inspect.getmembers(klass, inspect.ismethod)
        ds = inspect.getmembers(klass, inspect.isdatadescriptor)
        self.properties = {n: d for n, d in ds if isinstance(d, property)}
        self.attributes = {n: d for n, d in ds if not isinstance(d, property)}

        def _visit_FunctionDef_class(node):
            sd, ld, rt, ps, kw, va, decs = visit_FunctionDef(node)
            mi = FunctionInspector(
                name=node.name,
                shortDescription=sd,
                longDescription=ld,
                returns=rt,
                parameters=ps,
                keywords=kw,
                varargs=va,
                decorators=decs,
            )
            mi.function = getattr(self.klass, node.name, None)
            mi.doc = mi.function.__doc__
            mi.module = self.module
            mi.moduleName = self.moduleName
            mi.isStatic = "staticmethod" in mi.decorators
            # remove first argument if not static
            if not mi.isStatic:
                mi.parameters.pop(0)
            self.methods[node.name] = mi

        node_iter = ast.NodeVisitor()
        node_iter.visit_FunctionDef = _visit_FunctionDef_class

        # avoid builtin
        def is_builtin(obj):
            mn = obj.__module__
            return (mn in EXCLUDED_MODULES)

        if not is_builtin(self.klass):
            node = ast.parse(inspect.getsource(self.klass))
            node_iter.visit(node)

        self.mro = []
        for mro in inspect.getmro(self.klass)[1:]:
            if is_builtin(mro):
                break
            ci_mro = ClassInspector(mro)
            self.mro.append(ci_mro)
            self.attributesInherited.update(ci_mro.attributes)
            self.propertiesInherited.update(ci_mro.properties)
            self.methodsInherited.update(ci_mro.methods)

    def __repr__(self):
        return "<ClassInsp %s>" % self.name

    @property
    def init(self):
        return self.methods.get("__init__", None)

    @property
    def methodsAll(self):
        return dict(
            itertools.chain(
                six.iteritems(self.methods),
                six.iteritems(self.methodsInherited)))

    @property
    def methodsPublic(self):
        return {
            n: m
            for n, m in self.methodsAll.items()
            if n not in list(self.properties.keys() +
                             self.propertiesInherited.keys())
            and not n.startswith("_")
        }


def gcs(*instances):
    mros = (type(ins).mro() for ins in instances)
    mro = next(mros)
    common = set(mro).intersection(*mros)
    return next((x for x in mro if x in common), None)


def inspect_file(file_):
    """
    Retrieves the list of function names and the list of class names of a python
    source file

    :param file_: file path
    :type file_: [str, path]
    """
    r = open(str(file_), "r")
    t = ast.parse(r.read())
    return (
        [d.name for d in t.body if isinstance(d, ast.FunctionDef)],
        [d.name for d in t.body if isinstance(d, ast.ClassDef)],
        [
            _id(d.targets[0]) for d in t.body
            if isinstance(d, ast.Assign) and (
                hasattr(d.targets[0], "id") or hasattr(d.targets[0], "id"))
            and _id(d.targets[0])[0].isupper()
        ],
    )



# *- coding: utf-8 -*-
"""
json-schema validator classes

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext

import six
from jsonschema.validators import Draft6Validator
from jsonschema.validators import extend
from pyrsistent import pmap
from python_jsonschema_objects.validators import ValidationError
from python_jsonschema_objects.validators import converter_registry
from python_jsonschema_objects.validators import registry

from . import js_validators as _validators
from .schemas_loader import _load_schema

_ = gettext.gettext

# useful schemas shortcuts
SCH_STR = pmap({"type": "string"})
SCH_INT = pmap({"type": "integer"})
SCH_NUM = pmap({"type": "number"})
SCH_STR_ARRAY = pmap({"type": "array", "items": {"type": "string"}})
SCH_PATH = pmap({"type": "path"})
SCH_PATH_DIR = pmap({"type": "path", "isPathDir": True})
SCH_PATH_FILE = pmap({"type": "path", "isPathFile": True})
SCH_PATH_EXISTS = pmap({"type": "path", "isPathExisting": True})
SCH_DATE = pmap({"type": "date"})
SCH_TIME = pmap({"type": "time"})
SCH_DATETIME = pmap({"type": "datetime"})

NgoDraft04Validator = extend(
    Draft6Validator,
    validators={
        "$ref": _validators.ref_ngo_draft2,
        "extends": _validators.extends_ngo_draft1,
        "properties": _validators.properties_ngo_draft2,
    },
)
NgoDraft04Validator._setDefaults = False
NgoDraft04Validator.META_SCHEMA = _load_schema("ngo-draft-04")

DefaultValidator = NgoDraft04Validator


def convert_validate(value, schema):
    ret = value
    type_ = "enum" if "enum" in schema else schema.get("type", "object")

    converter = converter_registry(type_)
    if converter is not None:
        ret = converter(None, value, schema)

    if type_ == "array":
        assert isinstance(ret, (list, tuple))
        # validate items
        sch_items = schema.get("items", {})
        if sch_items:
            for i, e in enumerate(ret):
                ret[i] = convert_validate(e, sch_items)
        # validate length
        if "minItems" in schema and len(ret) < schema["minItems"]:
            raise ValidationError(
                "{1} has too few elements. Wanted {0}.".format(
                    schema["minItems"], value))
        if "maxItems" in schema and len(ret) > schema["maxItems"]:
            raise ValidationError(
                "{1} has too many elements. Wanted {0}.".format(
                    schema["maxItems"], value))
        # validate uniqueness
        if "uniqueItems" in schema and len(set(ret)) != len(ret):
            raise ValidationError(
                "{0} has duplicate elements, but uniqueness required".format(
                    value))

    for param, paramval in sorted(
            six.iteritems(schema), key=lambda x: x[0].lower() != "type"):
        validator = registry(param)
        if validator is not None:
            validator(paramval, ret, schema)

    return ret



# -*- coding: utf-8 -*-
"""
Unit tests for utilities

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
"""
from __future__ import print_function
from __future__ import unicode_literals

import collections
import pathlib
from builtins import str

import pytest
import six
from future.utils import text_to_native_str

from ngoschema import str_utils
from ngoschema import utils
from ngoschema.exceptions import InvalidValue


def test_resolve_import():
    function = utils.import_from_string("os.path.dirname")
    module = utils.import_from_string("past.builtins")
    klass = utils.import_from_string("decimal.Decimal")
    assert utils.fullname(klass) == "decimal.Decimal"
    with pytest.raises(InvalidValue) as e_info:
        utils.import_from_string("not.an.importable")

    method = utils.import_from_string(
        "jsonschema.validators.RefResolver.resolve")
    assert utils.fullname(
        method) == "jsonschema.validators.RefResolver.resolve"
    # tested on temporary file containing a nested class, import and fullname were working
    # nested_class = utils.import_from_string('ngoschema.example_nested_class.Human.Head')
    # assert utils.fullname(nested_class) == 'ngoschema.example_nested_class.Human.Head'


def test_utils():
    assert utils.is_sequence([1, 2])
    assert utils.is_sequence((1, 2))
    assert utils.is_sequence(collections.deque([1, 2]))
    assert not utils.is_sequence(set([1, 2]))

    assert utils.is_mapping({"a": 1})
    assert utils.is_mapping(collections.OrderedDict(a=1))
    assert not utils.is_mapping(1)
    assert not utils.is_mapping("a")

    assert utils.is_collection([1, 2])
    assert utils.is_collection(set([1, 2]))
    assert utils.is_collection((1, 2))
    assert utils.is_collection(collections.deque([1, 2]))
    assert utils.is_collection({"a": 1})
    assert utils.is_collection(collections.OrderedDict(a=1))
    assert not utils.is_collection(1)
    assert not utils.is_collection("a")

    newstring = str("foo")
    basestring_ = text_to_native_str("foo")
    unicodestr = str("foo")
    path = pathlib.Path("a/dummy/path")

    assert utils.is_basestring(basestring_) == six.PY2
    assert not utils.is_basestring(unicodestr)
    assert not utils.is_basestring(newstring)
    assert not utils.is_basestring(path)

    assert utils.is_string(basestring_)
    assert utils.is_string(unicodestr)
    assert utils.is_string(newstring)
    assert not utils.is_string(path)

    assert utils.is_pattern("{{ typical pattern }}")
    assert not utils.is_pattern("not a pattern")

    assert utils.is_expr("`a+b")
    assert not utils.is_expr("{{ typical pattern }}")


def test_string_utils():
    assert (str_utils.multiple_replace("boa kills, is it a boa", {
        "boa": "python",
        "kills": "rocks"
    }) == "python rocks, is it a python")

    assert str_utils.split_string("one,two|three",
                                  ",|") == ["one", "two", "three"]


if __name__ == "__main__":
    test_resolve_import()
    test_utils()
    test_string_utils()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import logging
import os
import pathlib

from future.utils import with_metaclass

from ngoschema.classbuilder import ProtocolBase
from ngoschema.deserializers import JsonDeserializer
from ngoschema.deserializers import YamlDeserializer
from ngoschema.schema_metaclass import SchemaMetaclass
from ngoschema.serializers import JsonSerializer
from ngoschema.serializers import YamlSerializer

_ = gettext.gettext

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("python_jsonschema_objects.classbuilder").setLevel(
    logging.INFO)

dirpath = os.path.dirname(os.path.realpath(__file__))


class Cookiecutter(with_metaclass(SchemaMetaclass, ProtocolBase)):
    schemaPath = os.path.join(dirpath, "schemas", "cookiecutter.json")


def test_json2yaml():

    # json -> yaml
    js = pathlib.Path(dirpath, "objects", "cc_ngoschema.json")
    cc_js = JsonDeserializer().load(js, objectClass=Cookiecutter)

    yml = js.with_name("cc_ngoschema_serialized.yaml")
    YamlSerializer().dump(
        {
            "default_context": cc_js.as_dict()
        }, yml, overwrite=True)
    cc_yml = YamlDeserializer().load(yml)
    cc_yml = cc_yml["default_context"]
    cc_yml = Cookiecutter(**cc_yml)
    assert cc_yml == cc_js


def test_yaml2json():
    # json -> yaml
    yml = pathlib.Path(dirpath, "objects", "cc_ngoschema.yaml")
    cc_yml = YamlDeserializer().load(yml)
    cc_yml = cc_yml["default_context"]
    cc_yml = Cookiecutter(**cc_yml)

    js = yml.with_name("cc_ngoschema_serialized.json")
    JsonSerializer().dump(cc_yml, js, overwrite=True)
    cc_js = JsonDeserializer().load(js, objectClass=Cookiecutter)
    assert cc_yml == cc_js


if __name__ == "__main__":
    test_json2yaml()
    test_yaml2json()



# *- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import copy
import ast
import inspect
"""
Copyright 2015, Wouter Bolsterlee

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or other materials
provided with the distribution.
Neither the name of the author nor the names of its contributors may be used to endorse
or promote products derived from this software without specific prior written
permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
Module to find out the qualified name of a class.
"""

__all__ = ["qualname"]

_cache = {}


class _Visitor(ast.NodeVisitor):
    def __init__(self):
        super(_Visitor, self).__init__()
        self.stack = []
        self.qualnames = {}

    def store_qualname(self, lineno):
        qn = ".".join(n for n in self.stack)
        self.qualnames[lineno] = qn

    def visit_FunctionDef(self, node):
        self.stack.append(node.name)
        self.store_qualname(node.lineno)
        self.stack.append("<locals>")
        self.generic_visit(node)
        self.stack.pop()
        self.stack.pop()

    def visit_ClassDef(self, node):
        self.stack.append(node.name)
        self.store_qualname(node.lineno)
        self.generic_visit(node)
        self.stack.pop()


def qualname(obj):
    """Find out the qualified name for a class or function."""

    # For Python 3.3+, this is straight-forward.
    if hasattr(obj, "__qualname__"):
        return obj.__qualname__

    # For older Python versions, things get complicated.
    # Obtain the filename and the line number where the
    # class/method/function is defined.
    try:
        filename = inspect.getsourcefile(obj)
    except TypeError:
        return obj.__qualname__  # raises a sensible error
    if inspect.isclass(obj):
        try:
            _, lineno = inspect.getsourcelines(obj)
        except (OSError, IOError):
            return obj.__qualname__  # raises a sensible error
    elif inspect.isfunction(obj) or inspect.ismethod(obj):
        if hasattr(obj, "im_func"):
            # Extract function from unbound method (Python 2)
            obj = obj.im_func
        try:
            code = obj.__code__
        except AttributeError:
            code = obj.func_code
        lineno = code.co_firstlineno
    else:
        return obj.__qualname__  # raises a sensible error

    # Re-parse the source file to figure out what the
    # __qualname__ should be by analysing the abstract
    # syntax tree. Use a cache to avoid doing this more
    # than once for the same file.
    qualnames = _cache.get(filename)
    if qualnames is None:
        with open(filename, "r") as fp:
            source = fp.read()
        node = ast.parse(source, filename)
        visitor = _Visitor()
        visitor.visit(node)
        _cache[filename] = qualnames = visitor.qualnames
        # add treatment to avoid problems with decorators
        lines = source.splitlines()
        orig_qualnames = copy.copy(visitor.qualnames)
        for i, v in orig_qualnames.items():
            i2 = i
            while True:
                if lines[i2 - 1].strip().startswith('@'):
                    i2 += 1
                    qualnames[i2] = v
                else:
                    break
        _cache[filename] = qualnames
    try:
        return qualnames[lineno]
    except KeyError:
        return obj.__qualname__  # raises a sensible error



# -*- coding: utf-8 -*-
"""
ngofile.exceptions
-----------------------
All exceptions used in the ngofile code base are defined here.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
from builtins import str

import jsonschema.exceptions

_ = gettext.gettext


class NgoSchemaException(Exception):
    """
    Base exception class. All ngofile-specific exceptions should subclass
    this class.
    """


class SchemaError(NgoSchemaException, jsonschema.exceptions.SchemaError):
    """
    Raised if an error is detected in a schema
    """


class InvalidValue(NgoSchemaException, jsonschema.exceptions.ValidationError,
                   ValueError):
    """
    Raised if an invalid value is detected in a validator
    """


class PropertyNotInSchema(NgoSchemaException):
    """
    Raisedfif property is not defined in Schema
    """

    def __init__(self, prop):
        message = (_(
            "[%(prop)s] does not exist in schema" % {"prop": str(prop)}))
        super(NgoSchemaException, self).__init__(message)


class PropertyUndefined(NgoSchemaException):
    """
    Raised property is not defined in Schema
    """

    def __init__(self, prop):
        message = _("[%(prop)s] is not defined" % {"prop": str(prop)})
        super(NgoSchemaException, self).__init__(message)



# *- coding: utf-8 -*-
"""
String utilities

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 02/01/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import re

from past.builtins import basestring

from .decorators import take_arrays

_ = gettext.gettext


def _multiple_replacer(replace_dict):
    replacement_function = lambda match: replace_dict[match.group(0)]
    pattern = re.compile(
        "|".join([re.escape(k) for k, v in list(replace_dict.items())]), re.M)
    return lambda string: pattern.sub(replacement_function, string)


def multiple_replace(string, key_values):
    """
    Replace efficiently multiple elements in a string according to a dictionary

    :param key_values: dictionary containing replacements
    :type key_values: dict
    :rtype: str
    """
    return _multiple_replacer(key_values)(string)


class CaseInsensitiveDict(dict):
    """
    Dictionary with keys insensitive to case
    """

    @classmethod
    def _k(cls, key):
        return key.lower() if isinstance(key, basestring) else key

    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveDict, self).__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key):
        return super(CaseInsensitiveDict, self).__getitem__(
            self.__class__._k(key))

    def __setitem__(self, key, value):
        super(CaseInsensitiveDict, self).__setitem__(
            self.__class__._k(key), value)

    def __delitem__(self, key):
        return super(CaseInsensitiveDict, self).__delitem__(
            self.__class__._k(key))

    def __contains__(self, key):
        return super(CaseInsensitiveDict, self).__contains__(
            self.__class__._k(key))

    def has_key(self, key):
        return super(CaseInsensitiveDict, self).has_key(self.__class__._k(key))

    def pop(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).pop(
            self.__class__._k(key), *args, **kwargs)

    def get(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).get(
            self.__class__._k(key), *args, **kwargs)

    def setdefault(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).setdefault(
            self.__class__._k(key), *args, **kwargs)

    def update(self, E={}, **F):
        super(CaseInsensitiveDict, self).update(self.__class__(E))
        super(CaseInsensitiveDict, self).update(self.__class__(**F))

    def _convert_keys(self):
        for k in list(self.keys()):
            v = super(CaseInsensitiveDict, self).pop(k)
            self.__setitem__(k, v)


@take_arrays(0)
def split_string(string, delimiters=" ", strip=True):
    """
    Split string with using several delimiters

    :param delimiters: string containing all delimiters
    :type delimiters: string
    :param strip: strip whitespaces around returned elements
    :rtype: list
    """
    if delimiters:
        specials = "[\^$.|?*+(){}"
        dels = ["\%s" % d if d in specials else d for d in delimiters]
        regex = "|".join([d for d in dels])
        return [w.strip() if strip else w for w in re.split(regex, string)]
    return [string]



# *- coding: utf-8 -*-
"""
Misc utilities

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 02/01/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import collections
import copy
import gettext
import importlib
import inspect
import re
import pathlib
import logging
import sys
import subprocess
from builtins import object
from builtins import str

from past.builtins import basestring
from ngofile.pathlist import PathList

from ._qualname import qualname
from .decorators import take_arrays
from .exceptions import InvalidValue

_ = gettext.gettext


class GenericRegistry(object):
    def __init__(self):
        self.registry = {}

    def register(self, name=None):
        def f(functor):
            self.registry[name
                          if name is not None else functor.__name__] = functor
            return functor

        return f

    def __call__(self, name):
        return self.registry.get(name)


class GenericModuleFileLoader(object):
    def __init__(self, subfolder_name):
        self.registry = {}
        self.subfolderName = subfolder_name

    def register(self, module, subfolder_name=None):
        m = importlib.import_module(module)
        subfolder_name = subfolder_name or self.subfolderName
        subfolder = pathlib.Path(
            m.__file__).with_name(subfolder_name).resolve()
        if subfolder.exists():
            if module not in self.registry:
                self.registry[module] = []
            self.registry[module].append(subfolder)
        return subfolder

    def find(self, name):
        """
        find first name/pattern in loader's pathlist

        :param name: path or pattern
        :rtype: path
        """
        name = name.replace('\\', '/')
        if '/' in name:
            module, path = name.split('/', 1)
            if module in self.registry:
                return PathList(*self.registry[module]).pick_first(path)
        all_paths = list(sum(self.registry.values(), []))
        return PathList(*all_paths).pick_first(name)


def gcs(*classes):
    """
    Return the greatest common superclass of input classes
    """
    mros = [x.mro() for x in classes]
    for x in mros[0]:
        if all([x in mro for mro in mros]):
            return x


def is_basestring(value):
    """
    Test if value is a basestring
    """
    return isinstance(value, basestring) and not isinstance(value, str)


def is_string(value):
    """
    Test if value is a string
    """
    return isinstance(value, (str, basestring))


def is_pattern(value):
    """
    Test if value is a pattern, ie contains {{ }} formatted content 
    """
    return is_string(value) and ("{{" in value or "{%" in value)


def is_expr(value):
    """
    Test if value is an expression and starts with `
    """
    return is_string(value) and value.strip().startswith("`")


def fullname(obj):
    if is_module(obj):
        return str(obj).split("'")[1]
    qn = getattr(obj, "__qualname__", None) or qualname(obj)
    mn = obj.__module__
    if mn is None or mn == str.__class__.__module__:  # avoid builtin
        return qn
    return mn + "." + qn


def import_from_string(value):
    """
    Imports a symbol from a string
    """
    poss = [m.start() for m in re.finditer(r"\.", "%s." % value)]
    # going backwards
    for pos in reversed(poss):
        try:
            m = value[0:pos]
            ret = importlib.import_module(m)
            for a in value[pos + 1:].split("."):
                if not a:
                    continue
                ret = getattr(ret, a, None)
                if not ret:
                    raise InvalidValue(
                        _("%s is not an importable object" % value))
            return ret
        except Exception as er:
            continue
    raise InvalidValue(_("%s is not an importable object" % value))


def is_module(value):
    """
    Test if value is a module
    """
    return inspect.ismodule(value)


def is_class(value):
    """
    Test if value is a class
    """
    return inspect.isclass(value)


def is_instance(value):
    """
    Test if value is an instance of a class
    """
    if getattr(value, "__class__"):
        return isinstance(value,
                          value.__class__) and not inspect.isclass(value)
    return False


def is_callable(value):
    """
    Test if value is a class
    """
    return is_instance(value) and hasattr(value, "__call__")


def is_static_method(value):
    """
    Test if value is a static method
    """
    return type(value) is staticmethod


def is_class_method(value):
    """
    Test if value is a class method
    """
    return type(value) is classmethod


def is_method(value,
              with_callable=True,
              with_static=True,
              with_class=True,
              with_method_descriptor=True):
    """
    Test if value is a method
    """
    if with_callable and is_callable(value):
        return True
    if with_static and is_static_method(value):
        return True
    if with_class and is_class_method(value):
        return True
    if with_method_descriptor and inspect.ismethoddescriptor(value):
        return True
    return inspect.ismethod(value)


def is_function(value, with_callable=True):
    """
    Test if value is a function
    """
    if with_callable and is_callable(value):
        return True
    return inspect.isfunction(value)


def is_imported(value):
    """
    Test if a symbol is importable/imported
    """
    return is_class(value) or is_method(value) or is_module(
        value) or is_function(value)


def is_importable(value):
    """
    Test if value is imported symbol or importable string
    """
    if is_string(value):
        try:
            value = import_from_string(value)
            return True
        except Exception as er:
            return False
    return is_imported(value)


def is_mapping(value):
    """
    Test if value is a mapping (dict, ordered dict, ...)
    """
    if isinstance(value, collections.Mapping):
        return True
    return False


def is_sequence(value):
    """
    Test if value is a sequence (list, tuple, deque)
    """
    if isinstance(value,
                  collections.Sequence) and not isinstance(value, basestring):
        return True
    if isinstance(value, collections.deque):
        return True
    return False


def is_collection(value):
    """
    Test if value is a list, set, tuple or dict
    """
    if is_mapping(value):
        return True
    if is_sequence(value):
        return True
    if isinstance(value, collections.Set):
        return True
    return False


def apply_through_collection(coll, func):
    """
    Generic method to go through a complex collection
    and apply a transformation function on elements
    """
    if is_mapping(coll):
        for k, v in coll.items():
            coll[k] = func(k, v)
            apply_through_collection(v, func)
    elif is_sequence(coll):
        for i, v in enumerate(coll):
            coll[i] = func(i, v)
            apply_through_collection(v, func)


def only_keys(icontainer, keys, recursive=False):
    """
    Keep only specific keys in a container

    :param icontainer: input container
    :param keys: keys to keep
    :type keys: list
    :param recursive: process container recursively
    """
    ocontainer = copy.deepcopy(icontainer)

    def delete_fields_not_of(container, fields, recursive):
        if is_mapping(container):
            to_del = fields.difference(set(container.keys()))
            left = fields.intersection(set(container.keys()))
            for k in to_del:
                del container[k]
            if recursive:
                for k in left:
                    delete_fields_not_of(container[k], fields, recursive)
        elif is_sequence(container):
            for i, v in enumerate(container):
                container[i] = delete_fields_not_of(v, fields, recursive)
        return container

    ocontainer = delete_fields_not_of(ocontainer, set(keys), recursive)
    return ocontainer


def but_keys(icontainer, keys, recursive=False):
    """
    Remove specific keys in a container

    :param icontainer: input container
    :param keys: keys to remove
    :type keys: list
    :param recursive: process container recursively
    """
    ocontainer = copy.deepcopy(icontainer)

    def delete_fields(container, fields, recursive):
        if is_mapping(container):
            to_del = fields.intersection(set(container.keys()))
            left = fields.difference(set(container.keys()))
            for k in to_del:
                del container[k]
            if recursive:
                for k in left:
                    delete_fields(container[k], fields, recursive)
        elif is_sequence(container):
            for i, v in enumerate(container):
                container[i] = delete_fields(v, fields, recursive)
        return container

    ocontainer = delete_fields(ocontainer, set(keys), recursive)
    return ocontainer


def process_collection(data, **opts):
    if "only_fields" in opts:
        rec = opts.get("fields_recursive", False)
        data = only_keys(data, opts["only_fields"], rec)
    if "but_fields" in opts:
        rec = opts.get("fields_recursive", False)
        data = but_keys(data, opts["only_fields"], rec)
    if "objectClass" in opts:
        return opts["objectClass"](**data)
    return data


def logging_call(popenargs,
                 logger=None,
                 stdout_log_level=logging.DEBUG,
                 stderr_log_level=logging.ERROR,
                 **kwargs):
    """
    Variant of subprocess.call that accepts a logger instead of stdout/stderr,
    and logs stdout messages via logger.debug and stderr messages via
    logger.error.

    inspired from code https://gist.github.com/1402841/231d4ae00325892ad30f6d9587446bc55c56dcb6
    """
    _logger = logger or logging.getLogger(__name__)
    out, err = subprocess.Popen(
        popenargs, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        **kwargs).communicate()
    #print out, err
    enc = sys.stdout.encoding or "cp850"
    if out:
        _logger.log(stdout_log_level, unicode(out, enc))
    if err:
        _logger.log(stderr_log_level, unicode(err, enc))



# *- coding: utf-8 -*-
""" schema validators to extend jsonschema library

_validators.py - created on 02/01/2018
author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import gettext
import pathlib
from builtins import str
from decimal import Decimal

import arrow
import six
from past.builtins import basestring
from python_jsonschema_objects import validators
from python_jsonschema_objects.validators import ValidationError
from python_jsonschema_objects.validators import converter_registry
from python_jsonschema_objects.validators import formatter_registry
from python_jsonschema_objects.validators import registry
from python_jsonschema_objects.validators import type_registry

from . import utils

_ = gettext.gettext

string_types = (basestring, str)
datetime_types = (datetime.datetime, arrow.Arrow)

NGO_TYPE_MAPPING = (
    ("path", string_types + (pathlib.Path, )),
    ("date", string_types + datetime_types + (datetime.date, )),
    ("time", string_types + datetime_types + (datetime.time, )),
    ("datetime", string_types + datetime_types),
)

validators.set_user_type_mapping(NGO_TYPE_MAPPING)

# validators
############


@registry.register()
def minimum(param, value, type_data):
    if value < param:
        raise ValidationError("{0} is less than {1}".format(value, param))


@registry.register()
def exclusiveMinimum(param, value, type_data):
    if value <= param:
        raise ValidationError("{0} is less than or equal to {1}".format(
            value, param))


@registry.register()
def maximum(param, value, type_data):
    if value > param:
        raise ValidationError("{0} is greater than {1}".format(value, param))


@registry.register()
def exclusiveMaximum(param, value, type_data):
    if value >= param:
        raise ValidationError("{0} is greater than or equal to {1}".format(
            value, param))


@registry.register()
def isPathDir(param, value, type_data):
    if value.is_dir() != param:
        raise ValidationError(
            "{0} is not the path of a directory".format(value))


@registry.register()
def isPathFile(param, value, type_data):
    if value.is_file() != param:
        raise ValidationError("{0} is not the path of a file".format(value))


@registry.register()
def isPathExisting(param, value, type_data):
    if value.exists() != param:
        raise ValidationError("{0} is not an existing path".format(value))


# type checkers
################


@type_registry.register(name="string")
def check_string_type(param, value, _):
    if not isinstance(value, string_types):
        raise ValidationError("{0} is not a string".format(value))


@type_registry.register(name="number")
def check_number_type(param, value, _):
    if not isinstance(value, six.integer_types + (float, Decimal)):
        raise ValidationError(
            "{0} is neither an integer nor a float".format(value))


@type_registry.register(name="path")
def check_path_type(param, value, _):
    if not isinstance(value, pathlib.Path):
        raise ValidationError("{0} is not a path".format(value))


@type_registry.register(name="date")
def check_date_type(param, value, _):
    if isinstance(value, datetime.date):
        return


@type_registry.register(name="time")
def check_time_type(param, value, _):
    if isinstance(value, datetime.time):
        return


@type_registry.register(name="datetime")
def check_datetime_type(param, value, _):
    if isinstance(value, arrow.Arrow):
        return


# converters
############


@converter_registry.register(name="string")
def convert_string(param, value, detail):
    if not utils.is_string(value) and utils.is_imported(value):
        return utils.fullname(value)
    return value


@converter_registry.register(name="array")
def convert_array(param, value, detail):
    if utils.is_sequence(value):
        return value
    return list(value)


@converter_registry.register(name="enum")
def convert_enum(param, value, details):
    if isinstance(value, int) and "enum" in details:
        if (value - 1) < len(details["enum"]):
            return details["enum"][value - 1]
    return value


date_fts = [
    "YYYY-MM-DD", "YYYY/MM/DD", "YYYY.MM.DD", "YYYY-MM", "YYYY/MM", "YYYY.MM"
]
alt_date_fts = [
    "DD-MM-YYYY",
    "DD/MM/YYYY",
    "DD.MM.YYYY",
    "MM-YYYY",
    "MM/YYYY",
    "MM.YYYY",
]


@converter_registry.register(name="date")
def convert_date_type(param, value, detail):
    if isinstance(value, datetime_types):
        return value.date()
    if isinstance(value, datetime.date):
        return value
    if isinstance(value, string_types):
        if value in ["now", "today"]:
            a = arrow.utcnow()
            return a.date()
        try:
            a = arrow.get(value, date_fts)
            if a.time() == datetime.time(0, 0):
                return a.date()
        except Exception:
            pass
        try:
            a = arrow.get(value, alt_date_fts)
            if a.time() == datetime.time(0, 0):
                return a.date()
        except Exception:
            pass
    raise ValidationError("{0} is not a date".format(value))


alt_time_fts = [
    "HH:mm:ssZZ",
    "HH:mm:ss ZZ",
    "HH:mm:ss",
    "HH:mm",
    "HH:mm:ss A",
    "HH:mm A",
]


@converter_registry.register(name="time")
def convert_time_type(param, value, _):
    if isinstance(value, datetime.time):
        return value
    if isinstance(value, datetime_types) and value.date() == datetime.date(
            1, 1, 1):
        return value.time()
    if isinstance(value, string_types):
        if value in ["now"]:
            a = arrow.utcnow()
            return a.time()
        try:
            a = arrow.get(value, alt_time_fts)
            if a.date() == datetime.date(1, 1, 1):
                return a.time()
        except Exception:
            pass
    raise ValidationError("{0} is not a time".format(value))


@converter_registry.register(name="datetime")
def convert_datetime_type(param, value, _):
    if isinstance(value, arrow.Arrow):
        return value
    if isinstance(value, datetime.datetime):
        return arrow.get(value)
    if isinstance(value, string_types):
        if value in ["now"]:
            a = arrow.utcnow()
            return a
    try:
        return arrow.get(value)
    except Exception:
        pass
    try:
        a_d = arrow.get(value, alt_date_fts)
        a_t = arrow.get(value, alt_time_fts)
        dt = datetime.datetime.combine(a_d.date(), a_t.time())
        return arrow.get(dt)
    except Exception:
        pass
    raise ValidationError("{0} is not a datetime".format(value))


@converter_registry.register(name="path")
def convert_path_type(param, value, _):
    if isinstance(value, pathlib.Path):
        return value
    if isinstance(value, string_types):
        return pathlib.Path(value)
    raise ValidationError("{0} is not a path".format(value))


# formatters
############


@formatter_registry.register(name="date")
def format_date(param, value, details):
    if "format" in details:
        frmt = details["format"]
        return value.strftime(frmt)
    return value.isoformat()


@formatter_registry.register(name="time")
def format_time(param, value, details):
    if "format" in details:
        frmt = details["format"]
        return value.strftime(frmt)
    return value.isoformat()


@formatter_registry.register(name="datetime")
def format_arrow(param, value, details):
    if "format" in details:
        frmt = details["format"]
        return value.format(frmt)
    return value



# -*- coding: utf-8 -*-
"""
Unit tests for schema validation

author: Cedric ROMAN
email: roman@numengo.com
licence: GNU GPLv3
created on 2018/05/06
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import json
import logging
import os
import pathlib

import pytest
import six
from jsonschema import Draft6Validator
from jsonschema import RefResolver
from python_jsonschema_objects.validators import ValidationError

from ngoschema import validators
from ngoschema.schemas_loader import load_schema_file
from ngoschema.schemas_loader import get_all_schemas_store
from ngoschema.resolver import ExpandingResolver
from ngoschema.validators import DefaultValidator

logging.basicConfig(level=logging.INFO)

test_dir = os.path.dirname(os.path.realpath(__file__))
defs_fp = os.path.join(test_dir, "schemas", "ngo-defs-draft-XX.json")
DEFS_URI, DEFS = load_schema_file(defs_fp)
MS_STORE = get_all_schemas_store()

resolver = ExpandingResolver(DEFS_URI, DEFS, MS_STORE)
orig_resolver = RefResolver(DEFS_URI, DEFS, MS_STORE)


def test_validate_extends():
    id, sch = resolver.resolve_by_name("FunctionArgument")
    # FunctionArgument extends 'metadata' which contains property 'title'
    id, sch2 = orig_resolver.resolve(id)

    instance = {
        "title": "RAISE ERROR: METADATA COMING FROM EXTENSION",
        "isArray": False,
        "refVariable": "#/variableGroups/1/Variables/variables/10/u1",
        "isDerivative": False,
        "cppName": "u1",
    }

    v = Draft6Validator(sch2)
    # title is a property from a class refered in "extended"
    with pytest.raises(Exception) as e_info:
        v.validate(instance)
        assert (
            e_info.message ==
            "Additional properties are not allowed (u'title' was unexpected)")

    v = Draft6Validator(sch)
    v.validate(instance)


def test_validate_subschemas():
    id, sch = resolver.resolve_by_name("ComponentDefinition")
    v_orig = Draft6Validator(sch, resolver=resolver)
    v_modif = DefaultValidator(sch, resolver=resolver)

    fp = os.path.join(test_dir, "schemas", "NgoPnodePhase.json")
    with open(fp, "rb") as f:
        instance = json.loads(f.read().decode("utf-8"))

    # even with an expanded schema, derived types are not resolved
    # derived types schemas are redefined with 'schema' field
    # errors should be due to missing properties (additionalProperties=False should raise those errors)
    # instances of errors should have a 'schema' field which override the actual schema with all missing properties
    errors = sorted(v_orig.iter_errors(instance), key=lambda e: e.path)
    for error in errors:
        # print(error.message)
        # print(list(error.path),list(error.schema_path))
        # pprint(error.instance)
        assert error.validator == "additionalProperties"
        assert error.instance.get("schemaUri")

    # NgoDraft4Validator handles derived types and should not raise errors
    v_modif.validate(instance)

    # the error is in the main expanded schema
    # a boolean has been turned to a string
    fp = os.path.join(test_dir, "schemas", "NgoPnodePhase_invalid1.json")
    with open(fp, "rb") as f:
        instance = json.loads(f.read().decode("utf-8"))

    errors = sorted(v_modif.iter_errors(instance), key=lambda e: e.path)
    assert len(errors) == 1
    error = errors[0]
    assert list(error.path) == ["initializeFromInputs"]
    if six.PY2:
        assert error.message == "u'true' is not of type u'boolean'"
    else:
        assert error.message == "'true' is not of type 'boolean'"

    # the error is further down in a subschema
    # the numerical value has been turned to a string
    fp = os.path.join(test_dir, "schemas", "NgoPnodePhase_invalid2.json")
    with open(fp, "rb") as f:
        instance = json.loads(f.read().decode("utf-8"))

    errors = sorted(v_modif.iter_errors(instance), key=lambda e: e.path)
    assert len(errors) == 1
    error = errors[0]
    assert list(error.path) == [
        "variableGroups",
        0,
        "variables",
        0,
        "literals",
        3,
        "numericalValue",
    ]
    if six.PY2:
        assert error.message == "u'4' is not of type u'integer'"
    else:
        assert error.message == "'4' is not of type 'integer'"


def test_convert_validate():
    with pytest.raises(ValidationError):
        validators.convert_validate([1, 2, 3], {
            "type": "array",
            "maxItems": 2
        })

    p = validators.convert_validate(__file__, validators.SCH_PATH_FILE)
    assert isinstance(p, pathlib.Path)

    with pytest.raises(ValidationError):
        validators.convert_validate(__file__, validators.SCH_PATH_DIR)

    d = validators.convert_validate("25/05/2018", validators.SCH_DATE)
    assert isinstance(d, datetime.date)

    e = validators.convert_validate(1, {"enum": ["A", "B", "C"]})
    assert e == "A"


if __name__ == "__main__":
    test_validate_extends()
    test_validate_subschemas()
    test_convert_validate()



# *- coding: utf-8 -*-
"""
Base class for loading objects from files

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import six

from future.utils import with_metaclass

from ngofile.list_files import list_files

from . import utils
from .classbuilder import ProtocolBase
from .schema_metaclass import SchemaMetaclass
from .transforms import ObjectTransform
from .deserializers import YamlDeserializer
from .deserializers import JsonDeserializer

_ = gettext.gettext


class ObjectLoader(with_metaclass(SchemaMetaclass, ProtocolBase)):
    """
    Class to load and translate models from files
    """

    schemaUri = "http://numengo.org/ngoschema/object_loader"
    deserializers = [JsonDeserializer, YamlDeserializer]
    primaryKey = "name"

    def __init__(self, **kwargs):
        ProtocolBase.__init__(self, **kwargs)

        self._oc = utils.import_from_string(str(
            self.objectClass)) if self.objectClass else None

        self._deserializers = [
            utils.import_from_string(str(ds)) for ds in self.deserializers
        ] if self.deserializers else []

        self._transforms = {}
        self._objects = {}

    def add_transformation(self, transfo):
        """
        Register an object transformation
        """
        transfo_ = transfo if hasattr(
            transfo, 'as_dict') else ObjectTransform(**transfo)
        if transfo_._from is None or transfo._to is None:
            raise ValueError(
                'transformation needs to have fully qualified from/to ' +
                'object classes')
        if issubclass(transfo_._from, self._oc):
            self._transforms[transfo_._to] = transfo_
        if issubclass(self._oc, transfo_._to):
            self._transforms[transfo_._from] = transfo_

    def _get_objects_from_data(self, data, objectClass, **opts):
        """
        Returns a list of objects found in data
        Can be overrided by subclasses to add specific treatments
        """
        obj = objectClass(**data)
        return [obj]

    def load_from_file(self, fp, **opts):
        """
        Load objects from a file
        Call protected method _process_data

        :type fp: path
        :param opts: options such as fromObjectClass
        """
        parsers = self._deserializers
        for p in parsers:
            try:
                data = p().load(fp, **opts)
                break
            except Exception as er:
                pass
        else:
            raise IOError(
                "Impossible to load %s with parsers %s.\n%s" % (fp, parsers))

        foc = opts.get('fromObjectClass') or self._oc
        try:
            if issubclass(self._oc, foc):
                objs = self._get_objects_from_data(data, self._oc, **opts)
            elif foc in self._transforms:
                tobjs = self._get_objects_from_data(data, foc)
                tf = self._transforms[foc]
                objs = [
                    tf.transform(tobj, objectClass=self._oc) for tobj in tobjs
                ]
        except Exception as er:
            raise IOError("Impossible to load %s from %s.\n%s" % (foc, fp, er))

        for obj in objs:
            ref = "%s#%s" % (fp, obj[str(self.primaryKey)])
            self._objects[ref] = obj
        return objs

    def load_from_directory(self,
                            src,
                            includes=["*"],
                            excludes=[],
                            recursive=False,
                            **opts):
        """
        Load from a search in a directory
       
        :type src: path
        """
        objs = []
        for fp in list_files(src, includes, excludes, recursive):
            try:
                objs += self.load_from_file(fp, **opts)
            except Exception as er:
                self.logger.warning(er)
        return objs

    def query(self, **kwargs):
        """
        Make a generator for a query in loaded objects
        """
        for path, obj in self._objects.items():
            for k in kwargs.keys():
                v = obj[k].for_json() if hasattr(obj[k],
                                                 "for_json") else obj[k]
                if v != kwargs[k]:
                    break
            else:
                yield obj

    def pick_first(self, **kwargs):
        """
        Pick first object corresponding to query
        """
        return next(self.query(**kwargs), None)

    @property
    def objects(self):
        """
        Return a list of all objects loaded
        """
        return self._objects.values()

    def __iter__(self):
        return six.itervalues(self._objects)

    def get(self, pk):
        """
        Return the first object with the corresponding primary key
        """
        return self.pick_first(**{str(self.primaryKey): pk})


# *- coding: utf-8 -*-
""" decorators

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 02/01/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import inspect
import sys
from pprint import pformat

import six
import wrapt
from python_jsonschema_objects.validators import ValidationError

from .exceptions import InvalidValue
from .validators import convert_validate

_ = gettext.gettext

# about decorators and why using wrapt
# https://hynek.me/articles/decorators/


def take_arrays(narg1=0, narg2=-1, flatten=False):
    """
    Decorator for functions/methods to take a list input and retrieve a list as
    a result

    :param narg1: index of argument to use as criteria
    :param narg2: index of an optional second argument (can be a list of same
                  size, or a scalar)
    :param flatten: flatten results
    :rtype: returns a list if the input argument is a list
    """

    def to_decorate(wrapped):
        sig = inspect.getargspec(wrapped)

        @wrapt.decorator
        def wrapper(method, instance, args, kwargs):
            # case first argument is an array
            if type(args[narg1]) in [list, set, tuple]:
                ret = []
                new_args = list(args)
                # only first argument is an array
                if (narg2 == -1 or not type(args[narg2]) in [list, set, tuple]
                        or len(args[narg2]) == 1):
                    for v in args[narg1]:
                        new_args[narg1] = v
                        ret.append(wrapped(*new_args, **kwargs))
                # both are list
                else:
                    # same size - call with zipped values
                    if len(args[narg1]) == len(args[narg2]):
                        for v1, v2 in zip(args[narg1], args[narg2]):
                            new_args[narg1] = v1
                            new_args[narg2] = v2
                            ret.append(wrapped(*new_args, **kwargs))
                    # different size
                    else:
                        raise InvalidValue(
                            _("arguments %s and %s must be of same size" %
                              (narg1, narg2)))
            else:
                # if the first one is not a string, ignore and call func
                return wrapped(*args, **kwargs)
            if flatten:
                return [item for sublist in ret for item in sublist]
            return ret

        decorated = wrapper(wrapped)

        doc = [wrapped.__doc__.strip()] if wrapped.__doc__ else []
        doc.append(
            "Argument '%s' can take an array, " % (sig.args[narg1]) +
            "and the function will perfom the operation on each element " +
            "and return the results as a list.")
        if narg2 != -1:
            doc.append(
                "Arguments '%s' and '%s'" %
                (sig.args[narg1], sig.args[narg2]) +
                " can take arrays of same size, and the function will perfom " +
                "the operation in parallel and return the results as a list.")
        if flatten:
            doc.append("The resulting list of lists if flattened.")

        wrapt.FunctionWrapper.__setattr__(decorated, "__doc__", "\n".join(doc))

        return decorated

    return to_decorate


def assert_arg(arg, schema):
    """
    Decorator to add a schema to validate a given argument against a json-schema

    :param arg: argument to convert/validate, can be position (start 0) or name
    :type arg: [str, int]
    :param schema: json-schema for the type
    :type schema: dict
    """

    def to_decorate(wrapped):
        # find argument in signature
        sig = inspect.getargspec(wrapped)
        if type(arg) is int:
            arg_i = arg
            arg_s = sig.args[arg]
        elif arg in sig.args:
            arg_i = sig.args.index(arg)
            arg_s = arg
        else:
            arg_i = None
            arg_s = arg

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            arg_i2 = arg_i if not instance and arg_i is not None else arg_i - 1
            args = list(args)
            try:
                if arg_s in kwargs:
                    kwargs[arg_s] = convert_validate(kwargs[arg_s], schema)
                elif type(arg_i2) is int and arg_i2 < len(args):
                    args[arg_i2] = convert_validate(args[arg_i2], schema)
                else:
                    raise Exception("unknown error")
            except ValidationError as er:
                if arg_s in kwargs:
                    raise ValidationError(
                        _("%s=%r is not valid. %s") % (arg_s, kwargs[arg_s],
                                                       er))
                elif type(arg_i) is int and arg_i2 < len(args):
                    raise ValidationError(
                        _("%s=%r is not valid. %s") % (arg_s, args[arg_i2],
                                                       er))
            return wrapped(*args, **kwargs)

        decorated = wrapper(wrapped)

        doc = (
            (wrapped.__doc__ or "").strip() + "\nArgument '%s' is " % arg +
            "automatically type converted and validated against this schema %s."
            % pformat(schema))

        wrapt.FunctionWrapper.__setattr__(decorated, "__doc__", doc)

        return decorated

    return to_decorate


def _format_call_msg(funcname, args, kwargs):
    """
    Format a call message
    """
    return "%s(%s)" % (
        funcname,
        ", ".join(["%r" % a for a in args] +
                  ["%s=%r" % (a, v) for a, v in kwargs.items()]),
    )


@wrapt.decorator
def log_exceptions(method, instance, args, kwargs):
    """
    log calls to a function/method with its arguments and the exceptions
    that might be raised.

    SHOULDN T BE USED FOR __init__, use log_init instead
    """
    # special case happening for getters/setters
    if instance is None:
        instance = args[0]
    try:
        if hasattr(instance, "logger"):
            instance.logger.debug("CALL " + _format_call_msg(
                "%r.%s" % (instance, getattr(method, __name__, 'unknown')),
                args, kwargs))
        return method(*args, **kwargs)
    except Exception as er:
        etype, value, trace = sys.exc_info()
        if hasattr(instance, "logger"):
            instance.logger.error("CALL " + _format_call_msg(
                "%r.%s" % (instance, getattr(method, __name__, 'unknown')),
                args, kwargs) + "\n\tERROR %s: %s" % (etype.__name__, value))
        try:
            six.reraise(etype, value, trace)
        finally:
            del trace  # to avoid circular refs


@wrapt.decorator
def log_init(init, instance, args, kwargs):
    """
    log init of instance and possible exceptions
    """
    instance.logger.info(
        _format_call_msg("INIT <%s>.__init__" % instance.__class__.__name__,
                         args, kwargs))
    try:
        return init(*args, **kwargs)
    except Exception as er:
        etype, value, trace = sys.exc_info()
        instance.logger.error("CALL " + _format_call_msg(
            "INIT <%s>.__init__" % instance.__class__.__name__, args, kwargs) +
                              "\n\tERROR %s: %s" % (etype.__name__, value))
        try:
            six.reraise(etype, value, trace)
        finally:
            del trace  # to avoid circular refs


def assert_prop(*args2check):
    """
    Assert some properties/attributes are defined in instance

    :param args2check: list of property/attribute to check as defined
    """

    def to_decorate(wrapped):
        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            for prop in args2check:
                if getattr(instance, prop, None) is None:
                    raise AttributeError(_("%s is not defined." % prop))
            return wrapped(*args, **kwargs)

        decorated = wrapper(wrapped)

        doc = (wrapped.__doc__ or "").strip()
        for prop in args2check:
            doc += "\nProperty %s is asserted to be defined before call." % prop
        wrapt.FunctionWrapper.__setattr__(decorated, "__doc__", doc)

        return decorated

    return to_decorate



# *- coding: utf-8 -*-
"""
Jinja2 serializer and custom filters

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import inspect
import logging
import pathlib
import re
import sys
from builtins import object
from builtins import str

import inflection
import jinja2

from . import utils
from .serializers import Serializer
from .serializers import serializer_registry

_ = gettext.gettext

templates_module_loader = utils.GenericModuleFileLoader('templates')

# default jinja2 environment instance
_default_jinja_env = None


def default_jinja2_env():
    """
    Return the default Jinja2 Environment with prefixed modules
    """
    global _default_jinja_env
    if _default_jinja_env is None:
        _default_jinja_env = ModulePrefixedJinja2Environment()
    return _default_jinja_env


regex_has_dot = re.compile(r"(\{\{\s*\w+\.)")


class templatedString(object):
    """
    Returns a templated string for a given context
    """

    def __init__(self, templated_str):
        self.templated_str = templated_str
        self.template = default_jinja2_env().from_string(templated_str)
        self.has_dot = regex_has_dot.search(templated_str) is not None

    def __call__(self, obj):
        # ctx = context.as_dict() if hasattr(context,'as_dict') else context
        if not hasattr(obj, "as_dict") or not self.has_dot:
            return self.template.render(obj)
        return self.template.render({"this": obj})


@serializer_registry.register()
class Jinja2Serializer(Serializer):
    logger = logging.getLogger(__name__)

    def __init__(self, template, environment=None):
        """ initialize with given template """
        self.jinja = environment or default_jinja2_env()
        self.template = template

    def dumps(self, obj, **opts):
        data = obj.as_dict() if hasattr(obj, "as_dict") else obj
        data = utils.process_collection(data, **opts)
        return self.jinja.get_template(self.template).render(data)


# ADDITIONAL FILTERS FROM INFLECTION

filters_registry = utils.GenericRegistry()


@filters_registry.register()
def camelize(string, uppercase_first_letter=True):
    __doc__ = inflection.camelize.__doc__
    return inflection.camelize(string, uppercase_first_letter)


@filters_registry.register()
def dasherize(word):
    __doc__ = inflection.dasherize.__doc__
    return inflection.dasherize(str(word))


@filters_registry.register()
def ordinal(number):
    __doc__ = inflection.ordinal.__doc__
    return inflection.ordinal(int(number))


@filters_registry.register()
def ordinalize(number):
    __doc__ = inflection.ordinalize.__doc__
    return inflection.ordinalize(int(number))


@filters_registry.register()
def parameterize(string, separator="-"):
    __doc__ = inflection.parameterize.__doc__
    return inflection.parameterize(str(string), separator)


@filters_registry.register()
def pluralize(word):
    __doc__ = inflection.pluralize.__doc__
    return inflection.pluralize(str(word))


@filters_registry.register()
def singularize(word):
    __doc__ = inflection.singularize.__doc__
    return inflection.singularize(str(word))


@filters_registry.register()
def tableize(word):
    __doc__ = inflection.tableize.__doc__
    return inflection.tableize(str(word))


@filters_registry.register()
def titleize(word):
    __doc__ = inflection.titleize.__doc__
    return inflection.titleize(str(word))


@filters_registry.register()
def transliterate(string):
    __doc__ = inflection.transliterate.__doc__
    return inflection.transliterate(str(string))


@filters_registry.register()
def underscore(word):
    __doc__ = inflection.underscore.__doc__
    return inflection.underscore(str(word))


class ModulePrefixedJinja2Environment(jinja2.Environment):
    logger = logging.getLogger(__name__ + ".DefaultJinja2Environment")

    def __init__(self, package_path="templates", **kwargs):
        # prepare prefixed loader with all loaded modules with package_path
        ms = {
            k: m
            for k, m in sys.modules.items()
            if m and '.' not in k and not k.startswith("_")
            and inspect.ismodule(m) and not inspect.isbuiltin(m)
        }
        to_load = {
            k: m
            for k, m in ms.items() if hasattr(m, "__file__")
            and pathlib.Path(m.__file__).with_name(package_path).is_dir()
        }
        loader = jinja2.PrefixLoader({
            k: jinja2.PackageLoader(utils.fullname(m), package_path)
            for k, m in to_load.items()
        })

        opts = {
            "trim_blocks": True,
            "lstrip_blocks": True,
            "keep_trailing_newline": True,
        }
        opts.update(kwargs)
        jinja2.Environment.__init__(self, loader=loader, **opts)

        # add filters
        for k, v in filters_registry.registry.items():
            self.filters[k] = v



# *- coding: utf-8 -*-
"""
Derived classbuilder from python-jsonschema-object for ngoschema specific
requirements

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 22/05/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import collections
import datetime
import gettext
import itertools
import copy
import inspect
import logging
import pathlib
from builtins import object

import arrow
import python_jsonschema_objects.classbuilder as pjo_classbuilder
import python_jsonschema_objects.literals as pjo_literals
import python_jsonschema_objects.pattern_properties as pjo_pattern_properties
import python_jsonschema_objects.util as pjo_util
import python_jsonschema_objects.validators as pjo_validators
import python_jsonschema_objects.wrapper_types as pjo_wrapper_types
from future.utils import text_to_native_str as native_str

from . import jinja2
from . import pjo_validators as ngo_pjo_validators
from . import utils
from .resolver import get_resolver
from .resolver import DEFAULT_MS_URI
from .config import ConfigLoader

_ = gettext.gettext

logger = pjo_classbuilder.logger

_NEW_TYPES = ngo_pjo_validators.NGO_TYPE_MAPPING

_default_builder = None


def get_builder(resolver=None):
    global _default_builder
    if _default_builder is None:
        _default_builder = ClassBuilder(resolver or get_resolver())
    else:
        base_uri = _default_builder.resolver.base_uri
        resolver = resolver or get_resolver(base_uri)
        _default_builder.resolver = resolver
    return _default_builder


def find_getter_setter_defv(propname, class_attrs):
    """
    Helper to retrieve getters/setters/default value in class attribute
    dictionary for a given property name
    """
    getter = None
    setter = None
    defv = None
    pn = propname
    gpn = "get_%s" % pn
    spn = "set_%s" % pn
    if pn in class_attrs:
        a = class_attrs[pn]
        if inspect.isfunction(a) or inspect.ismethod(a):
            logger.warning(
                pjo_util.lazy_format("{} will be overwritten", propname))
        elif inspect.isdatadescriptor(a):
            pass
        else:
            defv = a
    if gpn in class_attrs:
        a = class_attrs[gpn]
        if inspect.isfunction(a) or inspect.ismethod(a):
            getter = a
    if spn in class_attrs:
        a = class_attrs[spn]
        if inspect.isfunction(a) or inspect.ismethod(a):
            setter = a
    return getter, setter, defv


objects_config_loader = ConfigLoader()


class ProtocolBase(pjo_classbuilder.ProtocolBase):
    __doc__ = pjo_classbuilder.ProtocolBase.__doc__

    __class_attr_list__ = set()

    def __new__(cls, *args, **kwargs):
        if 'schemaUri' in kwargs:
            builder = get_builder()
            schemaUri = kwargs.pop("schemaUri")
            uri = pjo_util.resolve_ref_uri(builder.resolver.resolution_scope,
                                           schemaUri)
            if uri in builder.resolved:
                cls = builder.resolved[uri]
            else:
                uri, detail = builder.resolver.resolve(schemaUri)
                cls = builder.construct(uri, detail, (ProtocolBase, ))
            return cls(*args, **kwargs)
        new = super(ProtocolBase, cls).__new__
        if new is object.__new__:
            return new(cls)
        return new(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        # initialize Protocol Base
        self._name2attr = {}
        pjo_classbuilder.ProtocolBase.__init__(self, **kwargs)
        if getattr(self, '__attr_by_name__', False):
            for k, p in self.items():
                if not p:
                    continue
                if hasattr(p, "name"):
                    self._name2attr[str(p.name)] = (k, None)
                elif hasattr(p, "validate_items"):
                    for i, e in enumerate(p):
                        if hasattr(e, "name"):
                            self._name2attr[str(e.name)] = (k, i)

    def set_configfiles_defaults(self, overwrite=False):
        """
        Look for default values in objects_config_loader to initialize properties
        in the object.

        :param overwrite: overwrite values already set
        """
        defconf = objects_config_loader.get_values(self._fullname,
                                                   self._property_list)
        for k, v in defconf.items():
            if self._properties.get(k, None) or overwrite:
                try:
                    self.logger.debug(
                        "CONFIG SET %s.%s = %s" % (self.alt_repr(), k, v))
                    setattr(self, k, v)
                except Exception as er:
                    self.logger.error(er)

    @property
    def _fullname(self):
        return utils.fullname(self.__class__)

    @property
    def _property_list(self):
        return itertools.chain(self._properties.keys(),
                               self._extended_properties.keys())

    def alt_repr(self):
        return "<%s id=%i>" % (self.__class__.__name__, id(self))

    def __getattr__(self, name):
        """
        Allow getting class attributes, protected attributes and
        protocolBase attributes
        """
        if name in self.__class_attr_list__:
            collections.MutableMapping.__getattribute__(self, name)
        elif name.startswith("_"):
            return collections.MutableMapping.__getattribute__(self, name)
        else:
            if name in self._name2attr:
                n2a = self._name2attr[name]
                return getattr(self, n2a[0]) if n2a[1] is None else getattr(
                    self, n2a[0])[n2a[1]]
            return pjo_classbuilder.ProtocolBase.__getattr__(self, name)

    def __setattr__(self, name, val):
        """allow setting of protected attributes"""
        if name.startswith("_"):
            collections.MutableMapping.__setattr__(self, name, val)
        else:
            if name in self._name2attr:
                n2a = self._name2attr[name]
                if na2[1] is None:
                    pjo_classbuilder.ProtocolBase.__setattr__(
                        self, n2a[0], val)
                else:
                    getattr(self, n2a[0])[n2a[1]] = val
            else:
                pjo_classbuilder.ProtocolBase.__setattr__(self, name, val)

    def _set_prop_value(self, prop, value):
        """
        Set a property shorcutting the setter. To be used in setters
        """
        validator = self._properties.get(prop, None)
        if hasattr(validator, "validate"):
            validator(value)
            validator.validate()
        else:
            self._properties[prop] = value

    def _get_prop_value(self, prop):
        """
        Get a property shorcutting the setter. To be used in setters
        """
        if self._properties.get(prop, None) is not None:
            return self._properties[prop].for_json()


class ClassBuilder(pjo_classbuilder.ClassBuilder):
    """
    A modified ClassBuilder to build a class with SchemaMetaClass, to create
    properties according to schema, and associating with detected getter/setter
    or default values

    For a property PROP, the class builder will look for method called get_PROP
    or set_PROP to create the property with the setters/getters

    If a value of the same name is detected, it will attemp to use it as a
    default value.

    Additional pseudo-literal types are also handled (date, time, datetime,
    path). Those pseudo-literals will be properly deserialized/serialized
    and will provide all methods user would expect from standard python types.
        * date: datetime.date
        * time: datetime.time
        * arrow: arrow
        * path: pathlib.Path
    """

    def _build_pseudo_literal(self, nm, clsdata, parent):
        def __getattr__(self, name):
            """
            Special __getattr__ method to be able to use subclass methods
            directly on literal
            """
            if hasattr(self.__subclass__, name):
                return getattr(self._value, name)
            else:
                return pjo_literals.LiteralValue.__getattribute__(self, name)

        return type(
            native_str(nm),
            (pjo_literals.LiteralValue, ),
            {
                "__propinfo__": {
                    "__literal__": clsdata,
                    "__default__": clsdata.get("default"),
                },
                "__subclass__": parent,
                "__getattr__": __getattr__,
            },
        )

    def _construct(self, uri, clsdata, parent=(ProtocolBase, ), **kw):
        if clsdata.get("type") not in ("path", "date", "time", "datetime"):
            return pjo_classbuilder.ClassBuilder._construct(
                self, uri, clsdata, parent, **kw)

        typ = clsdata["type"]

        if typ == "path":
            self.resolved[uri] = self._build_pseudo_literal(
                uri, clsdata, pathlib.Path)
        if typ == "date":
            self.resolved[uri] = self._build_pseudo_literal(
                uri, clsdata, datetime.date)
        if typ == "time":
            self.resolved[uri] = self._build_pseudo_literal(
                uri, clsdata, datetime.time)
        if typ == "datetime":
            self.resolved[uri] = self._build_pseudo_literal(
                uri, clsdata, arrow.Arrow)

        return self.resolved[uri]

    def _build_object(self, nm, clsdata, parents, **kw):
        logger.debug(pjo_util.lazy_format("Building object {0}", nm))

        # To support circular references, we tag objects that we're
        # currently building as "under construction"
        self.under_construction.add(nm)

        # necessary to build type
        clsname = native_str(nm.split("/")[-1])

        props = dict()
        defaults = set()

        class_attrs = kw.get("class_attrs", {})

        # setup logger and make it a property
        if "logger" not in class_attrs:
            class_attrs["logger"] = logging.getLogger(clsname)

        # create a setter for logLevel
        if "logLevel" in clsdata.get("properties", {}):

            def set_logLevel(self, logLevel):
                level = logging.getLevelName(logLevel)
                self.logger.setLevel(level)

            class_attrs["set_logLevel"] = set_logLevel

        # we set class attributes as properties now, and they will be
        # overwritten if they are default values
        props.update(class_attrs)

        __object_attr_list__ = pjo_classbuilder.ProtocolBase.__object_attr_list__
        props["__object_attr_list__"] = __object_attr_list__

        props["__class_attr_list__"] = set(class_attrs.keys())

        properties = dict()

        # parent classes
        for ext in clsdata.get("extends", []):
            uri = pjo_util.resolve_ref_uri(self.resolver.resolution_scope, ext)
            if uri in self.resolved:
                base = self.resolved[uri]
                if not any([issubclass(p, base) for p in parents]):
                    parents = (base, ) + parents

        for p in reversed(parents):
            properties = pjo_util.propmerge(properties,
                                            getattr(p, "__propinfo__", {}))

        if "properties" in clsdata:
            properties = pjo_util.propmerge(
                properties, copy.deepcopy(clsdata["properties"]))

        name_translation = {}

        for prop, detail in properties.items():
            logger.debug(
                pjo_util.lazy_format("Handling property {0}.{1}", nm, prop))
            properties[prop]["raw_name"] = prop
            name_translation[prop] = prop.replace("@", "").replace("$", "")
            prop = name_translation[prop]

            # look for getter/setter/defaultvalue first in class definition
            getter, setter, defv = find_getter_setter_defv(prop, class_attrs)
            # look for missing getter/setter/defaultvalue in parent classes
            for p in reversed(parents):
                par_attrs = p.__dict__
                pgetter, psetter, pdefv = find_getter_setter_defv(
                    prop, par_attrs)
                getter = getter or pgetter
                setter = setter or psetter
                defv = defv or pdefv

            if defv is not None:
                detail["default"] = defv

            if detail.get("default") is None and detail.get("enum") is not None:
                detail['default'] = detail["enum"][0]

            if detail.get("default") is None and detail.get("type") == 'array':
                detail['default'] = []

            if detail.get("default") is not None:
                defaults.add(prop)

            if detail.get("type", None) == "object":
                uri = "{0}/{1}_{2}".format(nm, prop, "<anonymous>")
                self.resolved[uri] = self.construct(uri, detail,
                                                    (ProtocolBase, ))

                props[prop] = make_property(
                    prop,
                    {"type": self.resolved[uri]},
                    fget=getter,
                    fset=setter,
                    desc=self.resolved[uri].__doc__,
                )
                properties[prop]["type"] = self.resolved[uri]

            elif "type" not in detail and "$ref" in detail:
                ref = detail["$ref"]
                # TODO CRN: shouldn't we retrieve also the reference and construct from it??
                uri = pjo_util.resolve_ref_uri(self.resolver.resolution_scope,
                                               ref)
                logger.debug(
                    pjo_util.lazy_format("Resolving reference {0} for {1}.{2}",
                                         ref, nm, prop))
                if uri in self.resolved:
                    typ = self.resolved[uri]
                else:
                    typ = self.construct(uri, detail, (ProtocolBase, ))

                props[prop] = make_property(
                    prop, {"type": typ},
                    fget=getter,
                    fset=setter,
                    desc=typ.__doc__)
                properties[prop]["$ref"] = uri
                properties[prop]["type"] = typ

            elif "oneOf" in detail:
                potential = self.resolve_classes(detail["oneOf"])
                logger.debug(
                    pjo_util.lazy_format("Designating {0} as oneOf {1}", prop,
                                         potential))
                desc = detail["description"] if "description" in detail else ""
                props[prop] = make_property(
                    prop, {"type": potential},
                    fget=getter,
                    fset=setter,
                    desc=desc)

            elif "type" in detail and detail["type"] == "array":
                if "items" in detail and isinstance(detail["items"], dict):
                    if "$ref" in detail["items"]:
                        uri = pjo_util.resolve_ref_uri(
                            self.resolver.resolution_scope,
                            detail["items"]["$ref"])
                        typ = self.construct(uri, detail["items"])
                        propdata = {
                            "type":
                            "array",
                            "validator":
                            pjo_wrapper_types.ArrayWrapper.create(
                                uri, item_constraint=typ),
                        }
                    else:
                        uri = "{0}/{1}_{2}".format(nm, prop,
                                                   "<anonymous_field>")
                        try:
                            if "oneOf" in detail["items"]:
                                typ = pjo_classbuilder.TypeProxy([
                                    self.construct(uri + "_%s" % i,
                                                   item_detail)
                                    if "$ref" not in item_detail else
                                    self.construct(
                                        pjo_util.resolve_ref_uri(
                                            self.resolver.resolution_scope,
                                            item_detail["$ref"],
                                        ),
                                        item_detail,
                                    ) for i, item_detail in enumerate(detail[
                                        "items"]["oneOf"])
                                ])
                            else:
                                typ = self.construct(uri, detail["items"])
                            propdata = {
                                "type":
                                "array",
                                "validator":
                                pjo_wrapper_types.ArrayWrapper.create(
                                    uri,
                                    item_constraint=typ,
                                    addl_constraints=detail),
                            }
                        except NotImplementedError:
                            typ = detail["items"]
                            propdata = {
                                "type":
                                "array",
                                "validator":
                                pjo_wrapper_types.ArrayWrapper.create(
                                    uri,
                                    item_constraint=typ,
                                    addl_constraints=detail),
                            }

                    props[prop] = make_property(
                        prop,
                        propdata,
                        fget=getter,
                        fset=setter,
                        desc=typ.__doc__)
                elif "items" in detail:
                    typs = []
                    for i, elem in enumerate(detail["items"]):
                        uri = "{0}/{1}/<anonymous_{2}>".format(nm, prop, i)
                        typ = self.construct(uri, elem)
                        typs.append(typ)

                    props[prop] = make_property(
                        prop, {"type": typs}, fget=getter, fset=setter)

            else:
                desc = detail["description"] if "description" in detail else ""
                uri = "{0}/{1}".format(nm, prop)
                typ = self.construct(uri, detail)

                props[prop] = make_property(
                    prop, {"type": typ}, fget=getter, fset=setter, desc=desc)
        """
        If this object itself has a 'oneOf' designation, then
        make the validation 'type' the list of potential objects.
        """
        if "oneOf" in clsdata:
            klasses = self.resolve_classes(clsdata["oneOf"])
            # Need a validation to check that it meets one of them
            props["__validation__"] = {"type": klasses}

        props["__extensible__"] = pjo_pattern_properties.ExtensibleValidator(
            nm, clsdata, self)

        props["__prop_names__"] = name_translation

        props["__propinfo__"] = properties
        # required = set.union(*[p.__required__ for p in parents])
        required = set.union(
            *[getattr(p, "__required__", set()) for p in parents])

        if "required" in clsdata:
            for prop in clsdata["required"]:
                required.add(prop)

        invalid_requires = [
            req for req in required if req not in props["__propinfo__"]
        ]
        if len(invalid_requires) > 0:
            raise pjo_validators.ValidationError(
                "Schema Definition Error: {0} schema requires "
                "'{1}', but properties are not defined".format(
                    nm, invalid_requires))

        props["__required__"] = required
        props["__has_default__"] = defaults
        if required and kw.get("strict"):
            props["__strict__"] = True

        cls = type(clsname, tuple(parents), props)
        self.under_construction.remove(nm)

        return cls


def make_property(prop, info, fget=None, fset=None, fdel=None, desc=""):
    # flag to know if variable is readOnly
    RO = "readOnly" in info and info["readOnly"]
    RO_active = RO

    def getprop(self):
        self.logger.debug("GET %s.%s" % (self.alt_repr(), prop))
        if fget:
            try:
                RO_active = False
                setprop(self, fget(self))
            except Exception as er:
                RO_active = RO
                raise AttributeError(
                    _("Error getting property %s.\n%s" % (prop, er.message)))
        try:
            val = self._properties[prop]
            if hasattr(val, "_pattern"):
                evaluated = jinja2.templatedString(val._pattern)(self)
                val._value = evaluated
                val.validate()
            return val
        except KeyError:
            raise AttributeError(_("No attribute %s" % prop))

    def setprop(self, val):
        self.logger.debug("SET %s.%s=%s" % (self.alt_repr(), prop, val))
        if RO_active:
            raise AttributeError(_("'%s' is read only" % prop))

        if fset and self._properties[prop] is not None:
            # call the setter, and get the value stored in _properties
            fset(self, val)
            val = self._properties[prop]

        if isinstance(info["type"], (list, tuple)):
            ok = False
            errors = []
            type_checks = []

            for typ in info["type"]:
                if not isinstance(typ, dict):
                    type_checks.append(typ)
                    continue
                typ = next(t for n, t in pjo_validators.SCHEMA_TYPE_MAPPING +
                           pjo_validators.USER_TYPE_MAPPING
                           if typ["type"] == n)
                if typ is None:
                    typ = type(None)
                if isinstance(typ, (list, tuple)):
                    type_checks.extend(typ)
                else:
                    type_checks.append(typ)

            for typ in type_checks:
                if isinstance(val, typ):
                    ok = True
                    break
                elif hasattr(typ, "isLiteralClass"):
                    try:
                        validator = typ(val)
                    except Exception as e:
                        errors.append("Failed to coerce to '{0}': {1}".format(
                            typ, e))
                    else:
                        validator.validate()
                        ok = True
                        break
                elif pjo_util.safe_issubclass(typ, ProtocolBase):
                    # force conversion- thus the val rather than validator assignment
                    try:
                        val = typ(**pjo_util.coerce_for_expansion(val))
                    except Exception as e:
                        errors.append(
                            _("Failed to coerce to '%s': %s" % (typ, e)))
                    else:
                        val.validate()
                        ok = True
                        break
                elif pjo_util.safe_issubclass(typ,
                                              pjo_wrapper_types.ArrayWrapper):
                    try:
                        val = typ(val)
                    except Exception as e:
                        errors.append(
                            _(("Failed to coerce to '%s': %s") % (typ, e)))
                    else:
                        val.validate()
                        ok = True
                        break

            if not ok:
                errstr = "\n".join(errors)
                raise pjo_validators.ValidationError(
                    _("Object must be one of %s: \n%s" % (info["type"],
                                                          errstr)))

        elif info["type"] == "array":
            val = info["validator"](val)
            val.validate()

        elif pjo_util.safe_issubclass(info["type"],
                                      pjo_wrapper_types.ArrayWrapper):
            # An array type may have already been converted into an ArrayValidator
            val = info["type"](val)
            val.validate()
        elif hasattr(info["type"], "validate_items"):
            # THIS SHOULD NOT HAPPEN, SHOULD ALREADY TESTED TRUE JUST BEFORE
            raise Exception("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA")
            val = info["type"](val)
            val.validate()

        elif getattr(info["type"], "isLiteralClass", False) is True:
            if not isinstance(val, info["type"]):
                validator = info["type"](val)
                # handle case of patterns
                if utils.is_pattern(val):
                    validator._pattern = val
                # only validate if it s not a pattern
                else:
                    # it s not a pattern, remove
                    if hasattr(validator, "_pattern"):
                        delattr(validator, "_pattern")
                    validator.validate()
                if validator._value is not None:
                    # This allows setting of default Literal values
                    val = validator

        elif pjo_util.safe_issubclass(info["type"], ProtocolBase):
            if not isinstance(val, info["type"]):
                val = info["type"](**pjo_util.coerce_for_expansion(val))

            val.validate()

        elif isinstance(info["type"], pjo_classbuilder.TypeProxy):
            val = info["type"](val)

        elif isinstance(info["type"], pjo_classbuilder.TypeRef):
            if not isinstance(val, info["type"].ref_class):
                val = info["type"](**val)

            val.validate()

        elif info["type"] is None:
            # This is the null value
            if val is not None:
                raise pjo_validators.ValidationError(
                    _("None is only valid value for null"))

        else:
            raise TypeError(_("Unknown object type: '%s'" % (info["type"])))

        self._properties[prop] = val

    def delprop(self):
        self.logger.debug("DEL %s.%s" % (self.alt_repr(), prop))
        if prop in self.__required__:
            raise AttributeError(_("'%s' is required" % prop))
        else:
            if fdel:
                fdel(self)
            del self._properties[prop]

    return property(getprop, setprop, delprop, desc)



# *- coding: utf-8 -*-
"""
de/serializer meta class

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import codecs
import gettext
import json
import logging
import sys
from abc import ABCMeta
from abc import abstractmethod
from builtins import str

from future.utils import with_metaclass
from ruamel.yaml import YAML
from six import reraise as raise_

from . import decorators
from . import utils
from . import validators

_ = gettext.gettext


class Deserializer(with_metaclass(ABCMeta)):
    logger = logging.getLogger(__name__)

    @decorators.assert_arg(1, validators.SCH_PATH_FILE)
    def load(self, path, **opts):
        """
        Deserialize a file like object and returns the object

        :param path: file path containing the object
        :type path: path
        :param opts: dictionary of options, as protocol(=r) , encoding=(utf-8), objectClass=None
        """
        ptcl = opts.get("protocol", "r")
        enc = opts.get("encoding", "utf-8")
        self.logger.info(_("LOAD from file %s" % (path)))
        # with codecs.open(str(path), ptcl, enc) as f:
        #    return self.loads(f, **opts)
        with codecs.open(str(path.resolve()), ptcl, enc) as f:
            try:
                return self.loads(f.read(), **opts)
            except Exception as er:
                type, value, traceback = sys.exc_info()
                msg = "Problem occured loading file %s.\n%s" % (path, value)
                raise_(IOError, msg, traceback)

    @abstractmethod
    def loads(self, string, **opts):
        """
        Deserialize a string and returns the object

        :param string: data string
        :type string: string
        :param opts: dictionary of options, as protocol(=r) , encoding=(utf-8), objectClass=None
        """


deserializer_registry = utils.GenericRegistry()


@deserializer_registry.register()
class JsonDeserializer(Deserializer):
    logger = logging.getLogger(__name__ + ".JsonDeserializer")

    def loads(self, string, **opts):
        data = json.loads(string
                          # ,encoding=opts.get('encoding', 'utf-8')
                          )
        data = utils.process_collection(data, **opts)
        return data


@deserializer_registry.register()
class YamlDeserializer(Deserializer):
    logger = logging.getLogger(__name__ + ".YamlDeserializer")

    def __init__(self, **kwargs):
        # default, if not specfied, is 'rt' (round-trip)
        self._yaml = YAML(typ="safe", **kwargs)

    def loads(self, string, **opts):
        data = self._yaml.load(string)
        data = utils.process_collection(data, **opts)
        return data



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import logging
import os
import pathlib

from future.utils import with_metaclass

from ngoschema.classbuilder import ProtocolBase
from ngoschema.deserializers import JsonDeserializer
from ngoschema.schema_metaclass import SchemaMetaclass

_ = gettext.gettext

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("python_jsonschema_objects.classbuilder").setLevel(
    logging.INFO)

dirpath = os.path.dirname(os.path.realpath(__file__))


def test_transform():
    from ngoschema.transforms import ObjectTransform

    class Cookiecutter(with_metaclass(SchemaMetaclass, ProtocolBase)):
        schemaPath = os.path.join(dirpath, "schemas", "cookiecutter.json")

    class Project(with_metaclass(SchemaMetaclass, ProtocolBase)):
        schemaPath = os.path.join(dirpath, "schemas", "project.json")

    js = pathlib.Path(dirpath, "objects", "cc_ngoschema.json")
    cc_js = JsonDeserializer().load(js, objectClass=Cookiecutter)

    mtm_fp = pathlib.Path(dirpath, "transforms", "cookiecutter2project.mtm")
    mtm = JsonDeserializer().load(mtm_fp, objectClass=ObjectTransform)

    proj = mtm.transform_from(cc_js, objectClass=Project)
    # field equivalence transform
    assert proj.authorEmail == cc_js.email


if __name__ == "__main__":
    test_transform()



# *- coding: utf-8 -*-
"""
Utilities and classes to resolve schemas enabling inheritance.

The resolver expand a schema adding all the properties of the bases mentionned
in the 'extends' field

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created: 01/05/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import copy
import gettext

import dpath.util
from jsonschema.compat import urldefrag
from jsonschema.validators import RefResolver

from .schemas_loader import get_all_schemas_store

_ = gettext.gettext

CURRENT_DRAFT = "draft-04"


def uri_ngo(name, draft=CURRENT_DRAFT):
    return "http://numengo.org/%s/%s" % (draft, name)


DEFAULT_MS_URI = uri_ngo("schema")

_def_store = dict()


class ExpandingResolver(RefResolver):
    __doc__ = (""" 
    Resolver expanding the resolved document according to the 'extends' field
    (array of uri-reference). The returned document is a deep merge of all
    corresponding documents.
    The merge is done according to the order of 'extends' (properties can be
    overwritten).The original resolved document is merged at the end.
    """ + RefResolver.__doc__)

    _expanding = True

    def _expand(self, uri, schema):
        """ expand a schema to add properties of all definitions it extends 
        if a URI is given, it will be used to identify the schema in a cache store.
        If no resolver is given, use a resolver with local schema store, with the
        URI as referring document.
        
        :param schema: schema definition
        :type schema: dict 
        :param uri: schema uri
        :type uri: string
        :param resolver: resolver to use instead of default resolver.
        :rtype: dict
        """
        uri = self._urljoin_cache(self.resolution_scope, uri)

        doc_scope, frag = urldefrag(uri)

        ref = schema.get("$ref")

        if ref:
            ref = self._urljoin_cache(doc_scope, ref)
            uri_, schema_ = RefResolver.resolve(self, ref)
            if uri_ in _def_store:
                return _def_store[uri_]
            sch = self._expand(uri_, schema_)
            _def_store[uri_] = sch
            return sch

        if uri in _def_store:
            return _def_store[uri]

        schema_exp = {}

        extends = schema.get("extends", [])
        for ref in extends:
            ref = self._urljoin_cache(doc_scope, ref)
            uri_, schema_ = RefResolver.resolve(self, ref)
            sch = self._expand(uri_, schema_)
            dpath.util.merge(schema_exp, copy.deepcopy(sch))

        dpath.util.merge(schema_exp, copy.deepcopy(schema))

        extends = schema.get("extends", [])
        if extends:
            extends_ = []
            for e in extends:
                if e not in extends_:
                    extends_.append(e)
            schema_exp["extends"] = extends_

        #_def_store[uri] = schema_exp
        return schema_exp

    def resolve_from_url(self, url, expand=True):
        """
        Resolve a reference and returns its URI and the corresponding schema.

        :param url: reference to look for
        :type url: string
        :param expand: expand result if schema extends others
        :rtype: [string, dict]
        """
        schema = RefResolver.resolve_from_url(self, url)
        if expand:
            schema = self._expand(url, schema)
        return schema

    def resolve_by_name(self, name, expand=True):
        """
        Resolve a definition by name from the schema store.
        Return a tuple (uri,schema) or a list of tuple
        
        :param name: definition name
        :type name: string
        :rtype: tuple
        """
        res_scope = self.resolution_scope.split('#')[0]
        uris = list(self.store.keys())
        uris.remove(res_scope)
        uris.insert(0, res_scope)

        for uri in uris:
            s = self.store[uri]
            for p, d in dpath.util.search(
                    s.get("definitions", {}), "**/%s" % name, yielded=True):
                id = "%s#/definitions/%s" % (uri, p)
                if expand:
                    self.push_scope(uri)
                    d = self._expand(id, d)
                    self.pop_scope()
                return id, d


_resolver = None


def get_resolver(base_uri=DEFAULT_MS_URI):
    """
    Return a resolver set with the main loaded schema store
    with a base URI and the corresponding referred document.
    If no base_uri is defined, DEFAULT_MS_URI is used
    
    :param base_uri: base_uri to use for resolver
    :type base_uri: string
    """
    global _resolver
    base_uri, dummy = urldefrag(base_uri)
    ms = get_all_schemas_store()
    if base_uri not in ms:
        raise IOError(
            _("%s not found in loaded schemas (%s)" % (base_uri, ", ".join(
                ms.keys()))))
    referrer = ms[base_uri]
    if not _resolver or len(ms) != len(_resolver.store):
        _resolver = ExpandingResolver(base_uri, referrer, ms)
    if base_uri != _resolver.base_uri:
        _resolver.push_scope(base_uri)
    return _resolver



# -*- coding: utf-8 -*-
from .schemas_loader import load_module_schemas
load_module_schemas('ngoschema')

from .classbuilder import ProtocolBase
from .resolver import DEFAULT_MS_URI
from .resolver import get_resolver
from .schema_metaclass import SchemaMetaclass

__all__ = [
    "DEFAULT_MS_URI",
    "get_resolver",
    "ProtocolBase",
    "SchemaMetaclass",
]

__author__ = "Cédric ROMAN"
__email__ = "roman@numengo.com"
__version__ = "__version__ = '0.2.1'"



# *- coding: utf-8 -*-
"""
Utilities and classes to deal with schemas

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 02/01/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import gettext
import logging
import copy

from jsonschema.compat import iteritems
import python_jsonschema_objects.util as pjo_util

from . import decorators
from . import utils
from .classbuilder import get_builder
from .inspect_objects import FunctionInspector
from .resolver import get_resolver
#from .schemas_loader import load_module_schemas
from .schemas_loader import load_schema
from .schemas_loader import load_schema_file
from .validators import DefaultValidator

_ = gettext.gettext

logger = logging.getLogger(__name__)


class SchemaMetaclass(type):
    """
    Metaclass used for classes with schema
    """

    def __new__(cls, clsname, bases, attrs):
        """
        Process schema given by schemaPath if found and merges the schemas of bases

        When processing the schema dictionary, any harcoded initial value for
        schema is used.
        If a dictionary is found it will initialize the object define in schema
        """

        # base schema, should be overwritten
        schema = {}
        schemaUri = None
        # default resolver and builder
        builder = get_builder()
        resolver = builder.resolver
        if attrs.get("schema"):
            schemaUri, schema = load_schema(attrs["schema"])
            schema = resolver._expand(schemaUri, schema)
        elif attrs.get("schemaPath"):
            schemaUri, schema = load_schema_file(attrs["schemaPath"])
            schema = resolver._expand(schemaUri, schema)
        elif attrs.get("schemaUri"):
            schemaUri, schema = resolver.resolve(attrs["schemaUri"])
        if schema:
            # make a copy as building class will modify the dict and mess a lot things
            schema = copy.deepcopy(schema)
            # validate schema with its meta-schema
            metaschema = DefaultValidator.META_SCHEMA
            if schema.get("$schema"):
                ms_uri, metaschema = resolver.resolve(schema["$schema"])
            meta_validator = DefaultValidator(metaschema, resolver=resolver)
            # with hacked validator, can set a mode to set default values during
            # validation => schema will have its default values set
            def_bak = getattr(DefaultValidator, "_setDefaults", False)
            DefaultValidator._setDefaults = True
            meta_validator.validate(schema)
            DefaultValidator._setDefaults = def_bak

            logger.debug(_("creating <%s> with schema" % (clsname)))

            # reset resolver and builder to use the schemaUri as base
            resolver = get_resolver(schemaUri)
            builder = get_builder(resolver)
            # building inner definitions
            for nm, defn in iteritems(schema.get("definitions", {})):
                uri = pjo_util.resolve_ref_uri(schemaUri,
                                               "#/definitions/" + nm)
                builder.construct(uri, defn, attrs.get(nm, {}))
        else:
            schema["type"] = "object"

        # add some magic on methods defined in class
        # exception handling, argument conversion/validation, etc...
        for k, fn in attrs.items():
            if not (utils.is_method(fn) or utils.is_function(fn)):
                continue
            __add_logging__ = attrs.get("__add_logging__", False)
            __assert_props__ = attrs.get("__assert_props__", True)

            if __add_logging__ and k == "__init__":
                logger.debug(
                    _("decorate <%s>.__init__ with init logger" % (clsname)))
                fn = decorators.log_init(fn)

            # add argument checking
            if __assert_props__:
                fi = FunctionInspector(fn)
                for pos, p in enumerate(fi.parameters):
                    if p.type:
                        logger.debug(
                            _("decorate <%s>.%s " % (clsname, k) +
                              "with argument %i validity check." % pos))
                        fn = decorators.assert_arg(pos, p.type)(fn)

            # add exception logging
            if __add_logging__ and not k.startswith("__"):
                logger.debug(
                    _("decorate <%s>.%s with exception logger" % (clsname, k)))
                fn = decorators.log_exceptions(fn)

            attrs[k] = fn

        cls = builder.construct(
            clsname, schema, parent=bases, class_attrs=dict(attrs))
        if schemaUri is not None:
            builder.resolved[schemaUri] = cls
        return cls



# *- coding: utf-8 -*-
"""
Schema validators to extend jsonschema library

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
created on 20/05/2018
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import collections
import gettext
from builtins import str

from jsonschema._format import draft7_format_checker
from jsonschema.compat import iteritems
from jsonschema.exceptions import FormatError
from jsonschema.exceptions import RefResolutionError
from jsonschema.exceptions import ValidationError

_ = gettext.gettext


def _format_checker(validator):
    return validator.format_checker or draft7_format_checker


def extends_ngo_draft1(validator, extends, instance, schema):
    if validator.is_type(extends, "array"):
        for ref in extends:
            try:
                _format_checker(validator).check(ref, "uri-reference")
                scope, resolved = validator.resolver.resolve(ref)
            except FormatError as er:
                yield ValidationError(str(er), cause=er.cause)
            except RefResolutionError as er:
                yield ValidationError(str(er), cause=er.cause)


def properties_ngo_draft1(validator, properties, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    if "schema" in instance:
        scope, schema = validator.resolver.resolve(instance["schema"])
        validator.resolver.push_scope(scope)
        properties = schema.get("properties")

    for property, subschema in iteritems(properties):
        if getattr(validator, "_setDefault", False):
            if "default" in subschema and not isinstance(instance, list):
                instance.setdefault(property, subschema["default"])

        if property in instance:
            for error in validator.descend(
                    instance[property],
                    subschema,
                    path=property,
                    schema_path=property):
                yield error

    if "schema" in instance:
        validator.resolver.pop_scope()


def properties_ngo_draft2(validator, properties, instance, schema):
    if not validator.is_type(instance, "object"):
        return

    if "schemaUri" in instance:
        scope, schema = validator.resolver.resolve(instance["schemaUri"])
        validator.resolver.push_scope(scope)
        properties = schema.get("properties")

    for property, subschema in iteritems(properties):
        if getattr(validator, "_setDefault", False):
            if "default" in subschema and not isinstance(instance, list):
                instance.setdefault(property, subschema["default"])

        if property in instance:
            for error in validator.descend(
                    instance[property],
                    subschema,
                    path=property,
                    schema_path=property):
                yield error

    if "schemaUri" in instance:
        validator.resolver.pop_scope()


def ref_ngo_draft1(validator, ref, instance, schema):
    # override reference with schema defined in instance
    if isinstance(instance, collections.Iterable) and "schema" in instance:
        ref = instance["schema"]

    resolve = getattr(validator.resolver, "resolve", None)
    if resolve is None:
        with validator.resolver.resolving(ref) as resolved:
            for error in validator.descend(instance, resolved):
                yield error
    else:
        scope, resolved = validator.resolver.resolve(ref)
        validator.resolver.push_scope(scope)

        try:
            for error in validator.descend(instance, resolved):
                yield error
        finally:
            validator.resolver.pop_scope()


def ref_ngo_draft2(validator, ref, instance, schema):
    # override reference with schema defined in instance
    if isinstance(instance, collections.Iterable) and "schemaUri" in instance:
        ref = instance["schemaUri"]

    resolve = getattr(validator.resolver, "resolve", None)
    if resolve is None:
        with validator.resolver.resolving(ref) as resolved:
            for error in validator.descend(instance, resolved):
                yield error
    else:
        try:
            scope, resolved = validator.resolver.resolve(ref)
            validator.resolver.push_scope(scope)
        except RefResolutionError as error:
            yield ValidationError("%s. Resolution scope=%s" %
                                  (error, validator.resolver.resolution_scope))
            return

        try:
            for error in validator.descend(instance, resolved):
                yield error
        finally:
            # pass
            validator.resolver.pop_scope()



# *- coding: utf-8 -*-
"""
Utilities for config files

author: Cédric ROMAN (roman@numengo.com)
licence: GPL3
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import copy
import gettext
import logging
import appdirs
import pathlib

from builtins import object
from builtins import str

from backports import configparser2
from dpath.util import merge
from ngofile.pathlist import PathList

from .deserializers import Deserializer
from .str_utils import CaseInsensitiveDict

_ = gettext.gettext


def search_app_config_files(appname=None, appauthor=None, version=None):
    pl = PathList()
    cdirs = (appdirs.user_config_dir(appname, appauthor, version),
             appdirs.site_config_dir(appname, appauthor, version))
    for cdir in cdirs:
        if pathlib.Path(cdir).exists():
            pl.add(cdir)
    return list(pl.list_files(['*.cfg', '*.ini']))


class ConfigParser(Deserializer):
    logger = logging.getLogger(__name__)

    def loads(self, stream, **kwargs):
        config = configparser2.ConfigParser()
        config.read_string(stream)
        return config


class ConfigLoader(object):
    """
    Object to deal with multiple config files and merge them.

    When requesting a section, user retrieves a merged version of all section of
    same name in the different loaded files, including DEFAULT section, and
    inherited sections (using : as a separator between subsections).

    The returned dictionary is case insensitive. User can use get_defaults
    and provide a set of keys (with proper casing), it will return the
    corresponding dictionary with keys properly cased.
    """

    def __init__(self, *args, **kwargs):
        """
        User can supply a list of config files to merge
        """
        self._registry = {}
        self._sections = {}
        for f in args:
            self.add_config(f)

    def add_config(self, configFilepath):
        """
        Add a config file to registry
        """
        cfg = ConfigParser().load(configFilepath)
        self._registry[str(configFilepath)] = cfg

        for name, options in dict(cfg._sections).items():
            section = self._sections if name in self._sections else {}
            new_section = copy.copy(cfg._defaults)
            merge(new_section, section)
            merge(new_section, options)
            self._sections[name] = new_section

    def _section(self, name):
        parents = name.split(":")
        name = parents.pop(0)
        cfg = copy.copy(self._sections.get(name, {}))
        while parents:
            name += ":" + parents.pop(0)
            cfg.update(self._sections.get(name, {}))
        return CaseInsensitiveDict(cfg)

    def section(self, name):
        """
        Retrieve a section as a case insensitive dictionary, merging defaults
        and inherited sections.
        """
        return self._section(name)

    def get_values(self, sname, keys):
        """
        Method to retrieve the values in found in the given section of config
        files for the given set of keys

        :rtype: dict
        """
        section = self._section(sname)
        return {k: section[k] for k in keys if k in section}
